from radarpipeline.io.abc import Sampler
from typing import Any, Dict, List, Optional, Union
from random import sample
import pyspark.sql as ps
import logging


logger = logging.getLogger(__name__)


class UserSampler(Sampler):

    def __init__(self, config: Dict) -> None:
        super().__init__(config)

    def sample_uids(self, uid_list) -> None:
        if self.config['method'] == "fraction":
            fraction = self.config['config']['fraction']
            # sample fraction of the uids
            return self._sample_list(uid_list, round(len(uid_list) * fraction))
        elif self.config['method'] == "count":
            count = self.config['config']['count']
            # sample count of the uids
            return self._sample_list(uid_list, count)
        elif self.config['method'] == "userid":
            sampled_uids = self.config['config']['userids']
            for sampled_uid in sampled_uids:
                if sampled_uid not in uid_list:
                    logger.warning(f"User id {sampled_uid} not found in the data")
                    sampled_uids.remove(sampled_uid)
            return sampled_uids
        else:
            raise ValueError("Invalid method")

    def _sample_list(self, uid_list, number):
        return sample(uid_list, number)


class DataSampler(Sampler):
    def __init__(self, config: Dict) -> None:
        super().__init__(config)

    def sample_data(self, df: ps.DataFrame) -> Optional[ps.DataFrame]:
        if self.config['method'] == "fraction":
            fraction = self.config['config']['fraction']
            # sample fraction of the data
            return self._sample_data(df, fraction)
        elif self.config['method'] == "count":
            count = self.config['config']['count']
            # sample count of the data
            return self._sample_data(df, count / df.count())
        elif (self.config['method'] == "time"
              and not (type(self.config['config']) is list
                       and len(self.config['config']) != 1)):
            if len(self.config['config']) == 1:
                self.config['config'] = self.config['config'][0]
            starttime = self.config['config'].get('starttime', None)
            endtime = self.config['config'].get('endtime', None)
            time_col = self.config['config'].get('time_col', "value.time")
            # check if time_col is present in df
            if time_col not in df.columns:
                raise ValueError(f"Column {time_col} not found in the dataframe")
            return self._sample_data_by_time(df, starttime, endtime, time_col)
        elif self.config['method'] == "time" and type(self.config['config']) is list:
            sampled_dfs = []
            for time_config in self.config['config']:
                starttime = time_config.get('starttime', None)
                endtime = time_config.get('endtime', None)
                time_col = time_config.get('time_col', "value.time")
                # check if time_col is present in df
                if time_col not in df.columns:
                    raise ValueError(f"Column {time_col} not found in the dataframe")
                sampled_dfs.append(
                    self._sample_data_by_time(df, starttime, endtime, time_col))
            # combined sampled_dfs into one dataframe and deduplicate it
            return sampled_dfs[0].union(*sampled_dfs[1:]).distinct()
        else:
            raise ValueError("Invalid method")

    def _sample_data(self, df, fraction):
        return df.sample(fraction=fraction, withReplacement=True)

    def _sample_data_by_time(self, df, starttime, endtime, time_col):
        if endtime is None:
            return df.filter(df[f"`{time_col}`"] >= starttime)
        elif starttime is None:
            return df.filter(df[f"`{time_col}`"] < endtime)
        else:
            return df.filter(df[f"`{time_col}`"].between(starttime, endtime))



import unittest
from radarpipeline.project import Project
from radarpipeline.radarpipeline import radarpipeline
from strictyaml.exceptions import YAMLValidationError
import pathlib as pl
import os
import pandas as pd
from pandas.testing import assert_frame_equal


class TestCustomFeatureTabularize(unittest.TestCase):

    def assertIsFile(self, path):
        if not pl.Path(path).resolve().is_file():
            raise AssertionError("File does not exist: %s" % str(path))

    def test_run(self):
        # Test that the pipeline runs without error
        # This is a very basic test, but it's a start
        # Assert if pipeline is throwring an error
        try:
            radarpipeline.run(
                "tests/resources/test_yamls/config_with_custom_feature.yaml")
            raised = False
        except Exception:
            raised = True
        self.assertFalse(raised, 'Exception raised')
        project = Project(
            input_data="tests/resources/test_yamls/config_with_custom_feature.yaml")
        # Assert if output files are created
        self.output_dir = project.config['output']["config"]['target_path']
        path = pl.Path(os.path.join(
            self.output_dir,
            "tabularize_features_android_phone_battery_level.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output/tabular"
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_battery_level.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()
        path = pl.Path(
            os.path.join(
                self.output_dir,
                "tabularize_features_android_phone_step_count.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_step_count.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True))
        path.unlink()



import json
import logging
import os
from glob import glob
import gzip
import re
from typing import Any, Dict, List, Optional, Union
import concurrent.futures

import pyspark.sql as ps
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructField, StructType
from pyspark.sql.utils import IllegalArgumentException

from radarpipeline.common import constants
from radarpipeline.datalib import RadarData, RadarUserData, RadarVariableData
from radarpipeline.io.abc import DataReader, SchemaReader
from radarpipeline.io.sampler import UserSampler, DataSampler
from radarpipeline.common.utils import get_hash

import avro
from avro.datafile import DataFileReader, DataFileWriter
from avro.io import DatumReader, DatumWriter
from avro.schema import RecordSchema, Field, PrimitiveSchema, UnionSchema, Schema
from datetime import datetime

from collections import Counter
from functools import reduce, partial

logger = logging.getLogger(__name__)


class Schemas(object):
    def __init__(self, original_schema, original_schema_keys):
        self.original_schema = original_schema
        self.original_schema_hash = self._get_schema_hash(original_schema_keys)
        self.hashdict = {self.original_schema_hash: original_schema}

    def _get_schema_hash(self, schema_keys):
        return get_hash(schema_keys)

    def is_original_schema(self, schema_keys):
        return self._get_schema_hash(schema_keys) == self.original_schema_hash

    def is_schema_present(self, schema_keys):
        return self._get_schema_hash(schema_keys) in self.hashdict

    def is_schema_hash_present(self, schema_hash):
        return schema_hash in self.hashdict

    def get_schema(self, schema_keys=None):
        if schema_keys is None:
            return self.original_schema
        return self.hashdict[self._get_schema_hash(schema_keys)]

    def get_schema_by_hash(self, schema_hash):
        return self.hashdict[schema_hash]

    def add_schema(self, schema_keys, schema):
        schema_hash = self._get_schema_hash(schema_keys)
        self.hashdict[schema_hash] = schema


class Reader():
    '''
    Class for reading data from a file
    Reader(source_type : str, data_path: str, variables: Union[str, List])
    reader = Reader(...)
    reader.get_data(variables=Union[List, str])
    reader.get_user_data(user_id=..)
    '''
    def __init__(self, spark_session: ps.SparkSession,
                 config: Dict, required_data: Union[str, List[str]],
                 df_type: str = "pandas"):
        """_summary_

        Args:
            spark_session (ps.SparkSession): spark session instance
            config (Dict): Configuration data from the config.yaml file
            required_data (List[str]): List of required data
            df_type (str, optional): Type of dataframe format. Defaults to "pandas".
        """
        self.config = config
        self.source_type = self.config["input"]["data_format"]
        self.required_data = required_data
        self.df_type = df_type
        if self.config["configurations"]['user_sampling'] is None:
            self.user_sampler = None
        else:
            self.user_sampler = UserSampler(self.config["configurations"]
                                            ['user_sampling'])
        if self.config["configurations"]['data_sampling'] is None:
            self.data_sampler = None
        else:
            self.data_sampler = DataSampler(self.config["configurations"]
                                            ['data_sampling'])

        if self.source_type in ['csv', 'csv.gz']:
            self.reader_class = SparkCSVDataReader(spark_session, config,
                                                   required_data, df_type,
                                                   self.user_sampler, self.data_sampler)
        else:
            raise NotImplementedError("Only csv data type is supported for now")

    def read_data(self):
        self.data = self.reader_class.read_data()
        return self.data

    def get_data(self, variables: Union[List, str]) -> RadarData:
        return self.data.get_combined_data_by_variable(variables)

    def get_user_data(self, user_id: str) -> RadarData:
        return self.data.get_data_by_user_id(user_id)


class SparkCSVDataReader(DataReader):
    """
    Read CSV data from local directory using pySpark
    """

    def __init__(self, spark_session: ps.SparkSession,
                 config: Dict, required_data: Union[str, List[str]],
                 df_type: str = "pandas",
                 user_sampler: UserSampler = None,
                 data_sampler: DataSampler = None):
        super().__init__(config)
        self.source_formats = {
            # RADAR_OLD: uid/variable/yyyymmdd_hh00.csv.gz
            "RADAR_OLD": re.compile(r"""^[\w-]+/([\w]+)/
                                ([\d_]+.csv.gz|schema-\1.json)""", re.X),
            # RADAR_NEW: uid/variable/yyyymm/yyyymmdd.csv.gz
            "RADAR_NEW": re.compile(r"""^[\w-]+/([\w]+)/
                                    [\d]+/([\d]+.csv.gz$|schema-\1.json$)""", re.X),
            # RADAR_OLD: uid/questionnaire/QuestionaireName/yyyymm/yyyymmdd.csv.gz
            "RADAR_QUES": re.compile(r"""^[\w-]+/([\w]+)/([\w]+)/
                                    [\d]+/([\d]+.csv.gz$|schema-\1.json$)""", re.X)
        }
        self.required_data = required_data
        if isinstance(self.required_data, str):
            self.required_data = [self.required_data]
        self.df_type = df_type
        self.source_path = self.config['input']['config'].get("source_path", "")
        self.user_sampler = user_sampler
        self.data_sampler = data_sampler
        self.schema_reader = AvroSchemaReader()
        self.spark = spark_session
        self.unionByName = partial(DataFrame.unionByName, allowMissingColumns=True)

    def _get_source_type(self, source_path):
        """
        Returns the source type of the data
        """
        if source_path[-1] != "/":
            source_path = source_path + "/"
        for x in os.walk(source_path, topdown=False):
            for file in glob(os.path.join(x[0], '*.*')):
                for key, value in self.source_formats.items():
                    file_format = file.replace(source_path, "")
                    if re.match(value, file_format):
                        return key
        raise ValueError("Source path not recognized")

    def read_data(self) -> RadarData:
        """
        Reads RADAR data from local CSV files

        Returns
        -------
        RadarData
            A RadarData object containing all the read data
        """

        user_data_dict = {}

        if not isinstance(self.source_path, list):
            self.source_path = [self.source_path]
        for source_path_item in self.source_path:
            source_type = self._get_source_type(source_path_item)
            if source_type == "RADAR_OLD":
                logger.info("Reading data from old RADAR format")
                radar_data, user_data_dict = self._read_data_from_old_format(
                    source_path_item, user_data_dict)
            elif source_type == "RADAR_NEW" or source_type == "RADAR_QUES":
                logger.info("Reading data from new RADAR format")
                radar_data, user_data_dict = self._read_data_from_new_format(
                    source_path_item, user_data_dict)
        return radar_data

    def _filter_files_by_headers(self, data_files):
        file_dict = {}
        for file in data_files:
            with gzip.open(file, 'rb') as f:
                columns = f.readline().decode("utf-8").split(",")
                f.close()
            column_hash = get_hash(columns)
            if column_hash in file_dict:
                file_dict[column_hash].append(file)
            else:
                file_dict[column_hash] = [file]
        return file_dict

    def _read_variable_data_files(
        self,
        data_files: List[str],
        schema: Schemas = None,
        variable_name: Optional[str] = None
    ) -> RadarVariableData:
        """
        Reads data from a list of data files and returns a RadarVariableData object

        Parameters
        ----------
        data_files : List[str]
            List of data files to read
        schema : Optional[StructType]
            Schema to use for optimized reading

        Returns
        -------
        RadarVariableData
            A RadarVariableData object containing all the read data
        """
        dfs = []
        file_dict = self._filter_files_by_headers(data_files)
        if schema:
            for column_hash in file_dict.keys():
                if schema.is_schema_hash_present(column_hash):
                    df = self.spark.read.load(
                        file_dict[column_hash],
                        format="csv",
                        quote="\"",
                        escape="\"",
                        header=True,
                        schema=schema.get_schema_by_hash(column_hash),
                        enforceSchema="false",
                        encoding=constants.ENCODING,
                    )
                    dfs.append(df)
                else:
                    df = self.spark.read.load(
                        file_dict[column_hash],
                        format="csv",
                        quote="\"",
                        escape="\"",
                        header=True,
                        inferSchema="true",
                        encoding=constants.ENCODING,
                    )
                    inferred_schema = df.schema
                    schema.add_schema(df.columns, inferred_schema)
                    dfs.append(df)
        else:
            for column_hash in file_dict:
                df = self.spark.read.load(
                    file_dict[column_hash],
                    format="csv",
                    header=True,
                    inferSchema="true",
                    encoding=constants.ENCODING,
                )
                dfs.append(df)

        # Spark Join all the dfs
        # check if dfs are empty
        if len(dfs) == 0:
            # creating empty spark df
            df = self.spark.createDataFrame([], schema=schema.get_schema())
            variable_data = RadarVariableData(df, self.df_type)
        else:
            df = reduce(self.unionByName, dfs)
            variable_data = RadarVariableData(df, self.df_type,
                                              data_sampler=self.data_sampler)
        return variable_data

    def _read_data_from_old_format(self, source_path: str, user_data_dict: dict):
        uids = os.listdir(source_path)
        uids = self._remove_hidden_dirs(uids)
        if self.user_sampler is not None:
            uids = self.user_sampler.sample_uids(uids)

        def process_uid(uid):
            logger.info(f"Reading data for user: {uid}")
            variable_data_dict = {}
            for dirname in self.required_data:
                if dirname not in os.listdir(os.path.join(source_path, uid)):
                    continue
                logger.info(f"Reading data for variable: {dirname}")
                absolute_dirname = os.path.abspath(
                    os.path.join(source_path, uid, dirname)
                )
                data_files = [
                    os.path.join(absolute_dirname, f)
                    for f in os.listdir(absolute_dirname)
                    if f.endswith(".csv.gz")
                ]
                schema = None
                schema_dir = absolute_dirname
                schema_dir_base = self.schema_reader.get_schema_dir_base(schema_dir)
                if self.schema_reader.is_schema_present(schema_dir, schema_dir_base):
                    logger.info("Schema found")
                    schema = self.schema_reader.get_schema(schema_dir, schema_dir_base)
                else:
                    logger.info("Schema not found, inferring from data file")
                variable_data = self._read_variable_data_files(data_files, schema,
                                                               schema_dir_base)
                if variable_data.get_data_size() > 0:
                    variable_data_dict[dirname] = variable_data
            user_data_dict[uid] = RadarUserData(variable_data_dict, self.df_type)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(process_uid, uids)
        radar_data = RadarData(user_data_dict, self.df_type)
        return radar_data, user_data_dict

    def _read_data_from_new_format(self, source_path: str, user_data_dict: dict):
        # RADAR_NEW: uid/variable/yyyymm/yyyymmdd.csv.gz
        uids = os.listdir(source_path)
        uids = self._remove_hidden_dirs(uids)
        if self.user_sampler is not None:
            uids = self.user_sampler.sample_uids(uids)

        def process_uid(uid):
            # Skip hidden files
            if uid[0] == ".":
                return
            logger.info(f"Reading data for user: {uid}")
            variable_data_dict = {}
            for dirname in self.required_data:
                if not os.path.exists(os.path.join(source_path, uid, dirname)):
                    continue
                logger.info(f"Reading data for variable: {dirname}")
                data_files = []
                for date in os.listdir(os.path.join(source_path, uid, dirname)):
                    # Skip hidden files
                    if date[0] == ".":
                        continue
                    absolute_dirname = os.path.abspath(
                        os.path.join(source_path, uid, dirname, date))
                    data_files = data_files + [
                        os.path.join(absolute_dirname, f)
                        for f in os.listdir(absolute_dirname)
                        if f.endswith(".csv.gz")
                    ]
                schema = None
                schema_dir_base = dirname
                schema_dir = absolute_dirname
                if self.schema_reader.is_schema_present(schema_dir, schema_dir_base):
                    logger.info("Schema found")
                    schema = self.schema_reader.get_schema(schema_dir, schema_dir_base)
                else:
                    logger.info("Schema not found, inferring from data file")
                variable_data = self._read_variable_data_files(data_files, schema,
                                                               schema_dir_base)
                if variable_data.get_data_size() > 0:
                    variable_data_dict[dirname] = variable_data
            user_data_dict[uid] = RadarUserData(variable_data_dict, self.df_type)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(process_uid, uids)
        radar_data = RadarData(user_data_dict, self.df_type)
        return radar_data, user_data_dict

    def _remove_hidden_dirs(self, uids):
        return [uid for uid in uids if uid[0] != "."]


class AvroSchemaReader(SchemaReader):
    """
    Reads schema from local directory
    """

    def __init__(self) -> None:
        self.schema_dict = {}
        super().__init__(self.schema_dict)

    def get_schema_dir_base(self, schema_dir):
        return os.path.basename(schema_dir)

    def is_schema_present(self, schema_dir, schema_dir_base) -> bool:
        """
        Checks if schema is present in local directory

        Returns
        -------
        bool
            True if schema is present, False otherwise
        """
        if "/" in schema_dir_base:
            schema_dir_base = schema_dir_base.split("/")[0]
        schema_file = os.path.join(
            schema_dir, f"schema-{schema_dir_base}.json"
        )

        if os.path.exists(schema_file):
            return True
        return False

    def get_schema(self, schema_dir, schema_dir_base) -> StructType:

        if "/" in schema_dir_base:
            schema_dir_base = schema_dir_base.split("/")[0]
        if schema_dir_base in self.schema_dict:
            return self.schema_dict[schema_dir_base]
        else:
            schema, schema_keys = self._get_schema(schema_dir, schema_dir_base)
            schema_obj = Schemas(schema, schema_keys)
            self.schema_dict[schema_dir_base] = schema_obj
            return schema_obj

    def _get_schema(self, schema_dir, schema_dir_base) -> StructType:
        """
        Reads schema from local directory

        Returns
        -------
        StructType
            A StructType object defining the schema for pySpark
        """

        schema_file = os.path.join(
            schema_dir, f"schema-{schema_dir_base}.json"
        )
        schema_content = json.load(
            open(
                schema_file,
                "r",
                encoding=constants.ENCODING,
            )
        )
        avro_schema = avro.schema.parse(json.dumps(schema_content))
        schema_content_dict = self._recursive_schema_loader(avro_schema)

        schema, schema_keys = self._to_structtype(schema_content_dict)
        return schema, schema_keys

    def _add_new_schema(self, schema_dir_base, schema):
        self.schema_dict[schema_dir_base] = schema

    def _to_structtype(self, schema_dict):
        schema_fields = []
        schema_keys = []
        for key in schema_dict.keys():
            schema_fields.append(StructField(key, schema_dict[key], True))
            schema_keys.append(key)
        schema = StructType(schema_fields)
        return schema, schema_keys

    def _merge_dicts(self, dct1, dct2):
        return {**dct1, **dct2}

    def _recursive_schema_loader(self, record_schema, precursor="", schema_dict={}):
        """_summary_

        Args:
            record_schema (_type_): _description_
            precursor (str, optional): _description_. Defaults to "".
            schema_dict (dict, optional): _description_. Defaults to {}.
        """
        if isinstance(record_schema, RecordSchema):
            for f in record_schema.fields:
                schema_dict = self._merge_dicts(
                    schema_dict,
                    self._recursive_schema_loader(f, precursor, schema_dict)
                )
            return schema_dict
        elif isinstance(record_schema, Field):
            if isinstance(record_schema.type, RecordSchema):
                if precursor == "":
                    updated_precursor = record_schema.name
                else:
                    updated_precursor = precursor + "." + record_schema.name
                for f in record_schema.type.fields:
                    schema_dict = self._merge_dicts(
                        schema_dict,
                        self._recursive_schema_loader(f, updated_precursor, schema_dict)
                    )
                return schema_dict
            elif isinstance(record_schema.type, UnionSchema):
                if precursor == "":
                    updated_precursor = record_schema.name
                else:
                    updated_precursor = precursor + "." + record_schema.name
                is_record_schema_present = any(
                    isinstance(record_schema_ins, RecordSchema)
                    for record_schema_ins in record_schema.type.schemas
                )
                if is_record_schema_present:
                    for schema_instace in record_schema.type.schemas:
                        if isinstance(schema_instace, RecordSchema):
                            schema_dict = self._merge_dicts(
                                schema_dict,
                                self._recursive_schema_loader(
                                    schema_instace, updated_precursor, schema_dict
                                )
                            )
                else:
                    union_schema = self._resolve_union_schema(
                        record_schema.type.schemas
                    )
                    schema_dict[updated_precursor] = union_schema
                return schema_dict
            else:
                if precursor == "":
                    updated_precursor = record_schema.name
                else:
                    updated_precursor = precursor + "." + record_schema.name
                schema_dict = self._merge_dicts(
                    schema_dict,
                    self._recursive_schema_loader(
                        record_schema.type, updated_precursor, schema_dict)
                )
                return schema_dict

        elif isinstance(record_schema, PrimitiveSchema):
            schema_dict[precursor] = self._get_field(record_schema.type)
            return schema_dict
        else:
            return {}

    def _get_field(self, source_type: Union[str, Dict, List]) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        source_type : Union[str, Dict]
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if type(source_type) is dict:
            spark_data_type = self._get_data_type_from_dict(source_type)
        elif type(source_type) is list:
            spark_data_type = self._get_superior_type_from_list(source_type)
        else:
            spark_data_type = self._get_data_type_from_mapping(source_type)

        return spark_data_type

    def _resolve_union_schema(self, union_schemas: List[Schema]):
        list_type = []
        for schema in union_schemas:
            list_type.append(schema.type)
        return self._get_superior_type_from_list(list_type)

    def _handle_unknown_data_type(self, source_type: Union[str, Dict, List]) -> Any:
        """
        Handles unknown data types

        Parameters
        ----------
        source_type : Union[str, Dict]
            Data type to handle

        Returns
        -------
        Any
            A Spark data type
        """

        logger.warning(f"Unknown data type: {source_type}. Returning String type.")
        return constants.STRING_TYPE

    def _get_data_type_from_mapping(self, source_type: Union[str, Dict, List]) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        source_type : str
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if source_type in constants.DATA_TYPE_MAPPING:
            spark_data_type = constants.DATA_TYPE_MAPPING[source_type]
        else:
            spark_data_type = self._handle_unknown_data_type(source_type)

        return spark_data_type

    def _get_data_type_from_dict(self, source_type: Dict) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        source_type : Dict
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if "type" in source_type:
            return self._get_field(source_type["type"])
        else:
            return self._handle_unknown_data_type(source_type)

    def _get_superior_type_from_list(self, data_type_list: List[Any]) -> Any:
        """
        Resolves a list data type to a Spark data type

        Parameters
        ----------
        data_type_list : List
            List data type to resolve

        Returns
        -------
        Any
            A Spark data type
        """

        spark_data_type_list = data_type_list.copy()

        if "null" in spark_data_type_list:
            spark_data_type_list.remove("null")

        for index, source_type in enumerate(spark_data_type_list):
            if type(source_type) is dict:
                spark_data_type_list[index] = self._get_data_type_from_dict(source_type)
            elif source_type in constants.DATA_TYPE_MAPPING:
                spark_data_type_list[index] = constants.DATA_TYPE_MAPPING[source_type]
            else:
                spark_data_type_list[index] = self._handle_unknown_data_type(
                    source_type)

        if len(data_type_list) == 0:
            return constants.STRING_TYPE
        elif len(data_type_list) == 1:
            return spark_data_type_list[0]
        else:
            return self._get_superior_spark_type(spark_data_type_list)

    def _get_superior_spark_type(self, spark_data_type_list: List[Any]) -> Any:
        """
        Returns the superior Spark data type from a list of Spark data types

        Parameters
        ----------
        spark_data_type_list : List
            List of Spark data types

        Returns
        -------
        Any
            A Spark data type
        """

        spark_data_type_set = set(spark_data_type_list)

        # Conflicting types are datatypes which upon conversion can
        # lead to a loss of information
        are_conflicting_types = False
        are_integer_types = False
        are_float_types = False
        are_string_types = False
        are_boolean_types = False

        if len(spark_data_type_set.intersection(constants.INTEGER_TYPES)) > 0:
            are_integer_types = True

        if len(spark_data_type_set.intersection(constants.FLOATING_TYPES)) > 0:
            are_float_types = True

        if constants.STRING_TYPE in spark_data_type_set:
            are_string_types = True

        if constants.BOOLEAN_TYPE in spark_data_type_set:
            are_boolean_types = True

        if (
            int(are_integer_types)
            + int(are_float_types)
            + int(are_string_types)
            + int(are_boolean_types)
            > 1
        ):
            are_conflicting_types = True

        if not are_conflicting_types:
            if are_integer_types:
                if constants.LONG_TYPE in spark_data_type_set:
                    return constants.LONG_TYPE
                elif constants.INT_TYPE in spark_data_type_set:
                    return constants.INT_TYPE
                elif constants.SHORT_TYPE in spark_data_type_set:
                    return constants.SHORT_TYPE
                else:
                    return constants.BYTE_TYPE
            elif are_float_types:
                if constants.DOUBLE_TYPE in spark_data_type_set:
                    return constants.DOUBLE_TYPE
                else:
                    return constants.FLOAT_TYPE
            elif are_string_types:
                return constants.STRING_TYPE
            elif are_boolean_types:
                return constants.BOOLEAN_TYPE
        else:
            logger.warning(
                f"Conflicting types: {spark_data_type_list}. Returning String type."
            )
            return constants.STRING_TYPE



"""
The command-line interface for the radarpipeline
"""
import argparse
from radarpipeline import radarpipeline
import warnings
warnings.filterwarnings("ignore")


def main():
    parser = argparse.ArgumentParser(
        description="A CLI interface for radarpipeline"
    )
    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")

    # Subparser for 'run' command
    run_parser = subparsers.add_parser("run", help="Runs radarpipeline")
    run_parser.add_argument(
        "--config", "-f",
        help="Destination of config.yaml file"
    )
    run_parser.set_defaults(func=radarpipeline.run)

    # Subparser for 'validatååe' command
    validate_parser = subparsers.add_parser(
        "validate",
        help="Validate config file to run radarpipeline")

    validate_parser.add_argument(
        "--config", "-f",
        help="Destination of config.yaml file"
    )
    validate_parser.set_defaults(func=radarpipeline.validate)

    # Subparser for 'generate' command
    generate_parser = subparsers.add_parser(
        "generate",
        help="Generates a mock config file to run radarpipeline")

    generate_parser.add_argument(
        "--config", "-f",
        help="Destination of config.yaml file"
    )
    generate_parser.set_defaults(func=radarpipeline.generate_config)

    # Subparser for 'fetch' command
    fetch_parser = subparsers.add_parser("fetch", help="Fetch data using config file")
    fetch_parser.add_argument(
        "--config", "-f",
        help="Destination of config.yaml file"
    )
    fetch_parser.set_defaults(func=radarpipeline.fetch)

    # Subparser for 'convert' command
    convert_parser = subparsers.add_parser(
        "convert",
        help="Convert radar data to custom format"
    )

    # Mutually exclusive group for source path or config file
    source_group = convert_parser.add_mutually_exclusive_group()
    source_group.add_argument(
        "--source_path", "-s",
        help="Path to the source data to be converted"
    )
    source_group.add_argument(
        "--config", "-f",
        help="Destination of config.yaml file"
    )

    # Additional arguments for conversion
    convert_parser.add_argument(
        "--dest_path", "-d",
        help="Path where the converted data will be saved",
        default="./"
    )
    convert_parser.add_argument(
        "--variables", "-v",
        help="Variables to be converted",
        required=True,
        nargs='+'
    )
    convert_parser.add_argument(
        "--dest_format", "-df",
        help="Format to convert the data into",
        default="csv"
    )

    convert_parser.set_defaults(func=radarpipeline.convert)
    # Subparser for 'list' command
    list_parser = subparsers.add_parser("list", help="List available Pipelines")
    list_parser.set_defaults(func=radarpipeline.show_available_pipelines)

    args = parser.parse_args()
    if 'config' in args and args.config is not None:
        args.func(args.config)
    elif 'source_path' in args and args.source_path is not None:
        args.func(source_path=args.source_path, destination_path=args.dest_path,
                  variables=args.variables, data_format=args.dest_format)
    elif args.command == "generate":
        args.func()
    elif args.command == "list":
        pipelines = radarpipeline.show_available_pipelines()
        # print the list of pipelines in a readable table format for terminal
        # Pipeliness is a list of dict with keys name, url and description
        print("{:<30} {:<60} {:<40}".format("Name", "URL", "Description"))
        print("=" * 130)
        for pipeline in pipelines:
            print("{:<30} | {:<60} | {:<40}".format(
                pipeline["name"], pipeline["url"], pipeline["description"]))
    elif args.command is None:
        parser.print_help()
    else:
        raise ValueError("Invalid arguments")


if __name__ == "__main__":
    main()



import pyspark.sql as ps
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructField, StructType
from pyspark.sql.utils import IllegalArgumentException
from typing import Any, Dict, List, Optional, Union
import logging

logger = logging.getLogger(__name__)


class SparkEngine():
    """
    Read CSV data from local directory using pySpark
    """

    def __init__(self, spark_config: Dict = None):
        default_spark_config = {'spark_master': 'local',
                                'spark.executor.instances': 2,
                                'spark.driver.memory': '10G',
                                'spark.executor.cores': 4,
                                'spark.executor.memory': '10g',
                                'spark.memory.offHeap.enabled': True,
                                'spark.memory.offHeap.size': '20g',
                                'spark.driver.maxResultSize': '0',
                                'spark.log.level': "OFF"}
        self.spark_config = default_spark_config
        if spark_config is not None:
            self.spark_config.update(spark_config)

    def initialize_spark_session(self) -> ps.SparkSession:
        """
        Initializes and returns a SparkSession

        Returns
        -------
        SparkSession
            A SparkSession object
        """

        """
        Spark configuration documentation:
        https://spark.apache.org/docs/latest/configuration.html

        `spark.executor.instances` is the number of executors to
        launch for an application.

        `spark.executor.cores` is the number of cores to =
        use on each executor.

        `spark.executor.memory` is the amount of memory to
        use per executor process.

        `spark.driver.memory` is the amount of memory to use for the driver process,
        i.e. where SparkContext is initialized, in MiB unless otherwise specified.

        `spark.memory.offHeap.enabled` is to enable off-heap memory allocation

        `spark.memory.offHeap.size` is the absolute amount of memory which can be used
        for off-heap allocation, in bytes unless otherwise specified.

        `spark.driver.maxResultSize` is the limit of total size of serialized results of
        all partitions for each Spark action (e.g. collect) in bytes.
        Should be at least 1M, or 0 for unlimited.
        """
        if self.spark_config['spark_master'] == "local":
            self.spark = (
                SparkSession.builder.master("local[*]").appName("radarpipeline")
                .config('spark.executor.instances',
                        self.spark_config['spark.executor.instances'])
                .config('spark.executor.cores',
                        self.spark_config['spark.executor.cores'])
                .config('spark.executor.memory',
                        self.spark_config['spark.executor.memory'])
                .config('spark.driver.memory',
                        self.spark_config['spark.driver.memory'])
                .config('spark.memory.offHeap.enabled',
                        self.spark_config['spark.memory.offHeap.enabled'])
                .config('spark.memory.offHeap.size',
                        self.spark_config['spark.memory.offHeap.size'])
                .config('spark.driver.maxResultSize',
                        self.spark_config['spark.driver.maxResultSize'])
                .config('spark.log.level',
                        self.spark_config['spark.log.level'])
                .config("spark.serializer",
                        "org.apache.spark.serializer.KryoSerializer")
                .getOrCreate()
            )
        else:
            self.spark = (
                SparkSession.builder.master(
                    self.spark_config['spark_master']).appName("radarpipeline")
                .getOrCreate()
            )
        self.spark._jsc.setLogLevel(self.spark_config['spark.log.level'])
        self.spark.sparkContext.setLogLevel("OFF")
        # Enable Apache Arrow for optimizations in Spark to Pandas conversion
        self.spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        # Fallback to use non-Arrow conversion in case of errors
        self.spark.conf.set("spark.sql.execution.arrow.pyspark.fallback.enabled",
                            "true")
        self.spark.conf.set("spark.sql.session.timeZone", "UTC")
        # For further reading:
        # https://spark.apache.org/docs/3.0.1/sql-pyspark-pandas-with-arrow.html
        logger.info("Spark Session created")
        return self.spark

    def close_spark_session(self):
        self.spark.stop()



import os
from functools import reduce
from typing import Any, Dict, List
from urllib.parse import urlparse
import pathlib

import pyspark.sql as ps
import requests
from strictyaml import load, Map, Int, Str, Seq, Bool, Optional
from strictyaml import CommaSeparated, MapPattern
from dateutil import parser
import ntpath
import posixpath

from radarpipeline.common import constants
import pyspark.sql.functions as f
from pyspark.sql.types import TimestampType


def read_yaml(yaml_file_path: str) -> Dict[str, Any]:
    """
    Read yaml file and return a dict of the content

    Parameters
    ----------
    yaml_file_path : str
        Path to yaml file

    Returns
    -------
    Dict[str, Any]
        Dict of the content of the yaml file

    Raises
    ------
    ValueError
        If yaml file is not found or is not a yaml file, or is empty
    """

    if not os.path.exists(yaml_file_path):
        raise ValueError("Input file does not exist ")
    if not yaml_file_path.lower().endswith(".yaml"):
        raise ValueError("Input file is not a yaml file")
    if os.stat(yaml_file_path).st_size == 0:
        raise ValueError("Input file is empty")
    schema = get_yaml_schema()
    with open(yaml_file_path, "r", encoding=constants.ENCODING) as file:
        config = load(file.read(), schema).data
    return config


def combine_pyspark_dfs(dfs: List[ps.DataFrame]) -> ps.DataFrame:
    """
    Combine multiple pyspark dataframes of same schema into one

    Parameters
    ----------
    dfs : List[ps.DataFrame]
        List of pyspark dataframes

    Returns
    -------
    ps.DataFrame
        Combined pyspark dataframe
    """
    df_comb = reduce(
        lambda df1, df2: df1.unionByName(df2, allowMissingColumns=True), dfs
    )
    return df_comb


def is_valid_github_path(path: str) -> bool:
    """
    Check if the current path is a valid github path

    Parameters
    ----------
    path : str
        Path to check
    Returns
    -------
    bool
        True if the path is valid, False otherwise
    """

    parsed = urlparse(path)
    if parsed.scheme not in ["https", "http"] or parsed.netloc != "github.com":
        return False

    if requests.head(path, allow_redirects=True).status_code != 200:
        return False

    return True


def get_repo_name_from_url(url: str) -> str:
    """
    Return the name of the repository from the url
    Reference: https://stackoverflow.com/a/55137835/10307491

    Parameters
    ----------
    url : str
        Url of the repository

    Returns
    -------
    str
        Name of the repository
    """
    url = url.rstrip("/")
    last_slash_index = url.rfind("/")
    last_suffix_index = url.rfind(".git")
    if last_suffix_index < 0:
        last_suffix_index = len(url)

    if last_slash_index < 0 or last_suffix_index <= last_slash_index:
        raise Exception(f"Badly formatted url {url}")

    return url[last_slash_index + 1 : last_suffix_index]


def pascal_to_snake_case(s: str) -> str:
    """
    Convert a string from PascalCase to snake_case

    Parameters
    ----------
    s : str
        String to convert

    Returns
    -------
    str
        Converted string
    """
    return "".join(["_" + i.lower() if i.isupper() else i for i in s]).lstrip("_")


def get_yaml_schema() -> Map:
    schema = Map({
        "project": Map({
            "project_name": Str(),
            Optional("description"): Str(),
            Optional("version"): Str()
        }),
        "input": Map({
            "source_type": Str(),
            "config": MapPattern(Str(), Str()),
            "data_format": Str()
        }),
        "configurations": Map({
            "df_type": Str(),
            Optional("user_sampling"): Map({
                "method": Str(),
                "config": MapPattern(Str(), Seq(Str()) | Str()),
            }),
            Optional("data_sampling"): Map({
                "method": Str(),
                "config": MapPattern(Str(), Str()) | Seq(MapPattern(Str(), Str())),
            }),
        }),
        "features": Seq(Map({
            "location": Str(),
            Optional("branch", default='main'): Str(),
            "feature_groups": Seq(Str()),
            "feature_names": Seq(CommaSeparated(Str()))
        })),
        "output": Map({
            "output_location": Str(),
            "config": MapPattern(Str(), Str()),
            "data_format": Str(),
            "compress": Bool()
        }),
        Optional("spark_config"): Map({
            Optional("spark_master", default="local"): Str(),
            Optional("spark.executor.instances", default=4): Int(),
            Optional("spark.executor.cores", default=4): Int(),
            Optional("spark.executor.memory", default='10g'): Str(),
            Optional("spark.driver.memory", default='15g'): Str(),
            Optional("spark.memory.offHeap.enabled", default=True): Bool(),
            Optional("spark.memory.offHeap.size", default='20g'): Str(),
            Optional("spark.driver.maxResultSize", default='0'): Str(),
        }),
    })
    return schema


def get_absolute_path(path: str) -> str:
    """
    Returns the absolute path of the path
    Parameters
    ----------
    path: str
        Path to be converted to absolute path
    Returns
    -------
    str
        Absolute path of the path
    """
    if not os.path.isabs(path):
        pipeline_dir = pathlib.Path(__file__).parent.parent.parent.resolve()
        path = os.path.join(pipeline_dir, path)
    return path


def reparent(newparent, oldpath):
    '''when copying or moving a directory structure, you need to re-parent the
    oldpath.  When using os.path.join to calculate this new path, the
    appearance of a / root path at the beginning of oldpath, supplants the
    newparent and we don't want this to happen, so we need to make the oldpath
    root appear as a child of the newparent.

    :param: str newparent: the new parent location for oldpath (target)
    :param str oldpath: the path being adopted by newparent (source)

    :returns: (str) resulting adoptive path
    '''

    if oldpath[0] in (posixpath.sep, ntpath.sep):
        oldpath = '.' + oldpath
    return os.path.join(newparent, oldpath)


def get_write_file_attr(feature_name, output_dir, data_format, compression):
    """
    Returns the file name and path for writing the data

    Parameters
    ----------
    feature_name : str
        Name of the feature
    output_dir : str
        Output directory
    data_format : str
        Data format to be used
    compression : str
        Compression to be used

    Returns
    -------
    str
        File name
    str
        File path
    """
    if data_format == "csv":
        file_name = pascal_to_snake_case(feature_name) + ".csv"
        file_name += ".gz" if compression == "gzip" else ""
        file_path = os.path.join(output_dir, file_name)
    elif data_format == "parquet":
        file_name = pascal_to_snake_case(feature_name) + ".parquet"
        file_name += ".gz" if compression == "gzip" else ""
        file_path = os.path.join(output_dir, file_name)
    elif data_format == "pickle":
        file_name = pascal_to_snake_case(feature_name) + ".pkl"
        file_path = os.path.join(output_dir, file_name)
    else:
        raise ValueError(f"Invalid data format {data_format} specified \
            for spark writer")
    return file_path


def get_hash(array : List) -> int:
    """
    Returns the hash of the array

    Parameters
    ----------
    array : list
        List of values

    Returns
    -------
    str
        Hash of the array
    """
    return hash(tuple(array))


def preprocess_time_data(data):
    time_cols = ["value.time", "value.timeReceived", "value.dateTime",
                 "value.timeCompleted", "value.timeNotification"]
    for i, col in enumerate(time_cols):
        if col in data.columns:
            data = data.withColumn(col, data[f"`{col}`"].cast(TimestampType()))
            data.withColumn(col, f.from_unixtime(
                f.unix_timestamp(f"`{col}`")))
    return data


def convert_str_to_time(time):
    try:
        return parser.parse(time)
    except ValueError:
        raise ValueError(
            "Invalid value for the key: time. It should be a valid time format"
        )



from typing import List
import logging
import pandas as pd
import pyspark.sql.functions as f
from pyspark.sql.types import TimestampType

from radarpipeline.datalib.abc import Data
from radarpipeline.common.utils import preprocess_time_data
from radarpipeline.datatypes import DataType

logger = logging.getLogger(__name__)


class RadarVariableData(Data):
    """
    Class for reading data for a single variable of a single user
    """

    _data: DataType

    def __init__(self, data: DataType, df_type: str = "pandas",
                 data_sampler=None) -> None:
        self._data = data
        self.df_type = df_type
        self._preprocess_data()
        if data_sampler is not None:
            self._data = data_sampler.sample_data(self._data)

    def get_data(self) -> DataType:
        return self._data

    def set_data(self, data: DataType) -> None:
        self._data = data

    def get_data_keys(self) -> List[str]:
        return list(self._data.columns)

    def get_data_size(self) -> int:
        return int(self._data.count())

    def _preprocess_data(self) -> None:
        """
        Converts all time value columns to datetime format
        """
        try:
            self._data = preprocess_time_data(self._data)
        except ValueError:
            logger.warning("Unable to convert time columns to datetime format")



import logging

from radarpipeline.io.reader import Reader
from radarpipeline.project.sparkengine import SparkEngine

from typing import Dict

logger = logging.getLogger(__name__)


class CustomDataReader():
    def __init__(self, input_config, variables, source_type="local", data_format="csv",
                 df_type="pandas") -> None:
        self.variables = variables
        self.data_format = data_format
        self.source_type = source_type
        self.config = self.modify_config(input_config, data_format)
        self.sparkengine = SparkEngine()
        self.spark = self.sparkengine.initialize_spark_session()
        self.data_reader = Reader(self.spark, self.config, variables, df_type)

    def modify_config(self, input_config, data_format) -> Dict:
        """
        Modify the input configuration to include the variables of interest
        """
        config = {'input': {}, "configurations": {}}
        config['input']['config'] = input_config
        config['input']['data_format'] = data_format
        config['input']['source_type'] = self.source_type
        config['configurations']['df_type'] = "pandas"
        config['configurations']['user_sampling'] = None
        config['configurations']['data_sampling'] = None
        return config

    def read_data(self):
        return self.data_reader.read_data()

    def close_session(self):
        self.sparkengine.close_spark_session()



import importlib
import inspect
import logging
import os
import pathlib
import sys
from typing import Any, Dict, List, Union
import itertools

from git.exc import GitCommandError
from git.repo import Repo

from radarpipeline.common import utils
from radarpipeline.features import Feature, FeatureGroup
from radarpipeline.features.custom import Tabularize
from radarpipeline.io import PandasDataWriter, SparkDataWriter, Reader
from radarpipeline.io import SftpDataReader
from radarpipeline.project.validations import ConfigValidator
from radarpipeline.project.sparkengine import SparkEngine
from strictyaml import load, YAMLError

logger = logging.getLogger(__name__)


class Project:
    def __init__(self, input_data: Union[str, dict]) -> None:
        """
        Initialize the project with the data from the config file, or a config dict

        Parameters
        ---------
        input_data: Union[str, dict]
            Path to the config file or a dict containing the config
        """

        self.valid_input_formats = ["csv", "csv.gz"]
        self.valid_output_formats = ["csv", "pickle", "parquet"]
        self.custom_features = [Tabularize]
        self.input_data = self._resolve_input_data(input_data)
        self.feature_path = os.path.abspath(
            os.path.join("radarpipeline", "features", "features")
        )
        self.features = {}
        self.validate()
        self.feature_groups = self._get_feature_groups()
        self.total_required_data = self._get_total_required_data()
        if "spark_config" not in self.config:
            self.config["spark_config"] = {}
        self.spark_engine = SparkEngine(self.config['spark_config'])
        self.spark_session = self.spark_engine.initialize_spark_session()

    def close_spark_session(self):
        self.spark_engine.close_spark_session()

    def validate(self):
        self.config = self._get_config()
        self.validator = ConfigValidator(self.config, self.valid_input_formats,
                                         self.valid_output_formats)
        self.validator.validate()

    def _resolve_input_data(self, input_data) -> str:
        """
        Resolves the input_data variable and check
        if it is a local path or a remote github url.
        Returns:
            str: Path to the config file
        """
        # checking if input_data is a local path
        if isinstance(input_data, str):
            if os.path.exists(input_data):
                return input_data
            elif utils.is_valid_github_path(input_data):
                # get config.yaml from github
                try:
                    repo_name = utils.get_repo_name_from_url(input_data)
                    cache_dir = os.path.join(
                        os.path.expanduser("~"), ".cache", "radarpipeline", repo_name
                    )
                    if not os.path.exists(cache_dir):
                        Repo.clone_from(input_data, cache_dir)
                    input_data = os.path.join(cache_dir, "config.yaml")
                    return input_data
                except GitCommandError as err:
                    logger.error(f"Error while cloning the repo: {err}")
                    sys.exit(1)
            else:
                raise ValueError(f"File or URL does not exist: {input_data}")
        return input_data

    def _get_config(self) -> Dict[str, Any]:
        """
        Reads the config and returns it as a dictionary

        Returns
        -------
        Dict[str, Any]
            Config dictionary
        """

        if isinstance(self.input_data, str):
            try:
                config = utils.read_yaml(self.input_data)
            except YAMLError as err:
                logger.error(f"Error while reading config file: {err}")
                sys.exit(1)
        elif isinstance(self.input_data, dict):
            config = self.input_data
        else:
            raise ValueError("Wrong input data type. Should be yaml file path or dict.")

        logger.info("Config file read successfully")
        return config

    def _get_feature_groups(self) -> List[FeatureGroup]:
        """
        Get all the feature groups from the config

        Returns
        -------
        List[FeatureGroup]
            List of all the feature groups
        """

        features = self.config.get("features", [])
        feature_groups = []

        if "mock" in features:
            logger.info("Using mock features")
        else:
            for feature in features:
                feature_groups = feature_groups + self._get_feature_group(feature)
            feature_group_names = [
                feature_group.name for feature_group in feature_groups
            ]
            logger.info(f"Number of feature groups found: {len(list(feature_groups))}")
            logger.info(f"List of Feature groups: {feature_group_names}")

        return list(feature_groups)

    def _get_feature_group(self, feature: dict) -> List[FeatureGroup]:
        """
        Get feature group from filepath or name

        Parameters
        ----------
        feature: dict
            Feature to get from the specified location

        Returns
        -------
        List[FeatureGroup]
            List of feature group(s)
        """

        feature_group_classes = set()

        feature_location = feature["location"]
        req_feature_groups = feature["feature_groups"]

        if feature_location == "custom":
            all_feature_group_classes = [f(feature["feature_names"][0])
                                         for f in self.custom_features]
        else:
            # Get feature class from __init__.py file in feature_location
            all_feature_group_classes = self._get_feature_groups_from_filepath(
                feature_location
            )

        # Search feature_name in all_feature_classes
        for feature_group_class in all_feature_group_classes:
            if feature_group_class.name in req_feature_groups:
                feature_group_classes.add(feature_group_class)

        if len(feature_group_classes) == 0:
            raise ValueError(f"Feature not found: {feature}")

        return list(feature_group_classes)

    def _get_feature_groups_from_filepath(self, feature_location: str) -> List[Any]:
        """
        Gets the feature group classes from the filepath

        Parameters
        ----------
        feature_location: str
            Filepath of the feature group classes

        Returns
        -------
        List[Any]
            List of feature group classes
        """

        parent_dir = os.path.dirname(feature_location)
        base_name = os.path.basename(feature_location)

        sys.path.insert(1, parent_dir)

        # Import the feature module
        feature_module = importlib.import_module(base_name)

        # Get the feature group class
        feature_groups = set()
        for name, obj in inspect.getmembers(feature_module, inspect.isclass):
            if obj != Feature and obj != FeatureGroup:
                instance = obj()
                if isinstance(instance, FeatureGroup):
                    feature_groups.add(instance)

        return list(feature_groups)

    def _get_total_required_data(self) -> List[str]:
        """
        Get the total required data fromm the feature groups

        Returns
        -------
        List[str]
            List of all the required data
        """
        self.computable_feature_names = list(itertools.chain(*[
            f['feature_names'] for f in self.config["features"]]))
        total_required_data = set()
        for i, feature_group in enumerate(self.feature_groups):
            if self.computable_feature_names[i][0] == 'all' or feature_group.is_custom:
                total_required_data.update(feature_group.get_required_data())
            else:
                total_required_data.update(
                    feature_group.get_listed_required_data(
                        self.computable_feature_names[i]
                    )
                )
        logger.info(f"Total required data: {total_required_data}")
        return list(total_required_data)

    def read_data(self) -> None:
        """
        Read the data from the data source
        """
        self.fetch_data()
        if self.config["input"]["source_type"] == "local":
            datareader = Reader(
                self.spark_session,
                self.config,
                self.total_required_data,
                self.config["configurations"]["df_type"],
            )
            self.data = datareader.read_data()
        elif self.config["input"]["source_type"] == "mock":
            datareader = Reader(
                self.spark_session,
                self.mock_config,
                self.total_required_data,
                self.config["configurations"]["df_type"],
            )
            self.data = datareader.read_data()
        else:
            raise ValueError("Wrong data location")

    def fetch_data(self) -> None:
        """
        Fetch the data from the data source
        """
        if self.config["input"]["source_type"] == "mock":
            MOCK_URL = "https://github.com/RADAR-base-Analytics/mockdata"
            cache_dir = os.path.join(
                os.path.expanduser("~"), ".cache", "radarpipeline", "mockdata")
            if not os.path.exists(cache_dir):
                Repo.clone_from(MOCK_URL, cache_dir)
            mock_data_directory = os.path.join(cache_dir, "mockdata")
            self.mock_config = {
                "input": {
                    "config": {
                        "source_path": mock_data_directory},
                    "data_format": "csv"}}
            self.mock_config["configurations"] = self.config["configurations"]

        elif self.config["input"]["source_type"] == "sftp":
            sftp_data_reader = SftpDataReader(self.config["input"]["config"],
                                              self.total_required_data)
            root_dir = sftp_data_reader.get_root_dir()
            logger.info("Reading data from sftp")
            sftp_data_reader.read_sftp_data()
            self.config["input"] = {
                "source_type": "local",
                "data_format": "csv",
                "config": {
                    "source_path": root_dir
                },
            }
        elif self.config["input"]["source_type"] == "local":
            # check if source_path is a directory
            # if not, raise an error
            source_path = self.config["input"]["config"]["source_path"]
            if not os.path.exists(source_path):
                raise ValueError(f"Source path does not exist: {source_path}")
            if os.path.isdir(source_path):
                # check if there are any files in the directory
                if not os.listdir(source_path):
                    raise ValueError(f"Source path is empty: {source_path}")
        else:
            raise ValueError("Wrong data location")

    def compute_features(self) -> None:
        """
        Computes the features from the ingested data
        """
        for i, feature_group in enumerate(self.feature_groups):
            if self.computable_feature_names[i][0] == "all" or feature_group.is_custom:
                feature_names, feature_values = feature_group.get_all_features(
                    self.data
                )
            else:
                feature_names, feature_values = feature_group.get_listed_features(
                    self.computable_feature_names[i], self.data,
                )
            for feature_name, feature_value in zip(feature_names, feature_values):
                self.features[feature_name] = feature_value

    def export_data(self) -> None:
        """
        Exports the computed features to the specified location
        """
        df_type = self.config["configurations"]["df_type"]
        output_config = self.config["output"]
        if output_config['output_location'] == "local":
            if df_type == "pandas":
                writer = PandasDataWriter(
                    self.features,
                    output_config['config']["target_path"],
                    output_config["compress"],
                    output_config['data_format']
                )
            elif df_type == "spark":
                writer = SparkDataWriter(
                    self.features,
                    output_config['config']["target_path"],
                    output_config["compress"],
                    output_config['data_format']
                )
            else:
                raise ValueError("Wrong df_type")
        else:
            raise ValueError("Output location type not supported")
        writer.write_data()



from typing import Dict, List, Optional, Union

import pandas as pd

from radarpipeline.common import utils
from radarpipeline.datalib.abc import Data
from radarpipeline.datalib.radar_user_data import RadarUserData
from radarpipeline.datatypes import DataType


class RadarData(Data):
    """
    Class for reading the RADAR data
    """

    _data: Dict[str, RadarUserData]

    def __init__(self, data: Dict[str, RadarUserData], df_type: str = "pandas") -> None:
        self._data = data
        self.df_type = df_type

    def get_data(self) -> Dict[str, RadarUserData]:
        return self._data

    def set_data(self, data: Dict[str, RadarUserData]) -> None:
        self._data = data

    def get_data_type(self) -> str:
        return self.df_type

    def get_data_keys(self) -> List[str]:
        return list(self._data.keys())

    def get_data_size(self) -> int:
        return len(self._data.items())

    def _get_data_by_key(self, key: str) -> Optional[RadarUserData]:
        return self._data.get(key, None)

    def _get_all_user_ids(self) -> List[str]:
        """
        Get all user ids in the data of the RADAR data

        Returns
        -------
        List[str]
            The list of all user ids in the data of the RADAR data
        """

        return self.get_data_keys()

    def get_combined_data_by_variable(
        self, variables: Union[str, List[str]],
        return_dict: bool = False
    ) -> Union[DataType, List[DataType]]:
        """
        Returns the combined data of the RADAR data for the given variables

        Parameters
        ----------
        variables : Union[str, List[str]]
            The variable(s) to get the data for

        Returns
        -------
        Union[
            DataType,
            List[DataType]
        ]
            The combined data of the RADAR data for the given variables
        """
        is_only_one_var, variables = self._normalize_variables(variables)
        variable_dict = self._collect_variable_data(variables)
        return self._combine_and_return_data(variable_dict, is_only_one_var,
                                             return_dict)

    def _normalize_variables(self, variables: Union[str, List[str]]):
        is_only_one_var = isinstance(variables, str)
        if is_only_one_var:
            variables = [variables]
        return is_only_one_var, variables

    def _collect_variable_data(self, variables: List[str]) -> Dict[str, List[DataType]]:
        variable_dict = {}
        for user_id in self._get_all_user_ids():
            user_data = self._get_data_by_key(user_id)
            if user_data:
                for var in variables:
                    if var in user_data._get_all_variables():
                        var_data = user_data.get_data_by_variable(var)
                        if var_data:
                            variable_dict.setdefault(var, []).append(
                                var_data.get_data())
        return variable_dict

    def _combine_and_return_data(
        self, variable_dict: Dict[str, List[DataType]],
        is_only_one_var: bool, return_dict: bool
    ) -> Union[DataType, List[DataType]]:
        variable_data_list = []
        variable_data_dict = {}

        for var, data_list in variable_dict.items():
            if data_list:
                combined_df = utils.combine_pyspark_dfs(data_list)
                if self.df_type == "pandas":
                    combined_df = combined_df.toPandas()
                if return_dict:
                    variable_data_dict[var] = combined_df
                else:
                    variable_data_list.append(combined_df)

        if is_only_one_var:
            if return_dict:
                return variable_data_dict
            elif variable_data_list:
                return variable_data_list[0]
            else:
                raise ValueError(
                    f"No data found for the variable {list(variable_dict.keys())}")
        return variable_data_dict if return_dict else variable_data_list

    def get_data_by_user_id(
        self, user_ids: Union[str, List[str]]
    ) -> Union[RadarUserData, List[RadarUserData]]:
        """
        Returns the data of the RADAR data for the given user ids

        Parameters
        ----------
        user_ids : Union[str, List[str]]
            The user id(s) to get the data for

        Returns
        -------
        Union[
            RadarUserData,
            List[RadarUserData]
        ]
            The data of the RADAR data for the given user ids
        """

        is_only_one_var = False
        if isinstance(user_ids, str):
            is_only_one_var = True
            user_ids = [user_ids]

        all_user_ids = self._get_all_user_ids()
        user_data_list = []

        for user_id in user_ids:
            if user_id in all_user_ids:
                user_data = self._get_data_by_key(user_id)
                if user_data is not None:
                    user_data_list.append(user_data)

        if is_only_one_var:
            return user_data_list[0]
        else:
            return user_data_list

    def get_variable_data(
            self, variables: Union[List, str]) -> Union[DataType, List[DataType]]:
        return self.get_combined_data_by_variable(variables)

    def get_user_data(self, user_id: str) -> Union[DataType, List[DataType]]:
        return self.get_data_by_user_id(user_id)



import paramiko
from paramiko import AgentKey, RSAKey, DSSKey
from paramiko import SSHException, AuthenticationException

from typing import Any, Dict, List, Optional, Union
import json
import logging
import os
from radarpipeline.common.utils import reparent
import posixpath

from contextlib import contextmanager
from stat import S_IMODE, S_ISDIR, S_ISREG
import warnings

logger = logging.getLogger(__name__)


class WTCallbacks(object):
    '''an object to house the callbacks, used internally'''
    def __init__(self):
        '''set instance vars'''
        self._flist = []
        self._dlist = []
        self._ulist = []

    def file_cb(self, pathname):
        '''called for regular files, appends pathname to .flist

        :param str pathname: file path
        '''
        self._flist.append(pathname)

    def dir_cb(self, pathname):
        '''called for directories, appends pathname to .dlist

        :param str pathname: directory path
        '''
        self._dlist.append(pathname)

    def unk_cb(self, pathname):
        '''called for unknown file types, appends pathname to .ulist

        :param str pathname: unknown entity path
        '''
        self._ulist.append(pathname)

    @property
    def flist(self):
        '''return a sorted list of files currently traversed

        :getter: returns the list
        :setter: sets the list
        :type: list
        '''
        return sorted(self._flist)

    @flist.setter
    def flist(self, val):
        '''setter for _flist '''
        self._flist = val

    @property
    def dlist(self):
        '''return a sorted list of directories currently traversed

        :getter: returns the list
        :setter: sets the list
        :type: list
        '''
        return sorted(self._dlist)

    @dlist.setter
    def dlist(self, val):
        '''setter for _dlist '''
        self._dlist = val

    @property
    def ulist(self):
        '''return a sorted list of unknown entities currently traversed

        :getter: returns the list
        :setter: sets the list
        :type: list
        '''
        return sorted(self._ulist)

    @ulist.setter
    def ulist(self, val):
        '''setter for _ulist '''
        self._ulist = val


class ConnectionException(Exception):
    """Exception raised for connection problems

    Attributes:
        message  -- explanation of the error
    """

    def __init__(self, host, port):
        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, host, port)
        self.message = 'Could not connect to host:port.  %s:%s'


class CredentialException(Exception):
    """Exception raised for credential problems

    Attributes:
        message  -- explanation of the error
    """

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        Exception.__init__(self, message)
        self.message = message


class HostKeysException(Exception):
    '''raised when a problem with HostKeys is encountered'''
    pass


class SftpConnector():
    """
    Class for reading data from SFTP
    """
    def __init__(self, config_dict: Dict, variables: List[str]):
        """
        Parameters
        ----------
        config_dict : Dict
            Dictionary containing configuration for SFTP
        variables : List
            List of variables to read
        root_dir : str, optional
        """

        self.config_dict = config_dict
        self.variables = variables

    def connect(self):
        """
        Connect to SFTP server
        """
        username = self.config_dict.get('sftp_username')
        host = self.config_dict.get('sftp_host')
        password = self.config_dict.get('sftp_password')
        port = self.config_dict.get('sftp_port', 22)
        private_key = self.config_dict.get('sftp_private_key')
        self._tconnect = {'username': username, 'password': password, 'hostkey': None,
                          'pkey': None}
        self._sftp_live = False
        self._transport = None
        self._transport = paramiko.Transport((host, port))
        self._transport.banner_timeout = 200
        self._transport.auth_timeout = 200
        self._transport.use_compression(False)
        private_key_pass = None
        self._set_authentication(password, private_key, private_key_pass)
        self._transport.connect(**self._tconnect)
        self._sftp_connect()

    def _set_authentication(self, password, private_key, private_key_pass):
        '''Authenticate the transport. prefer password if given'''
        if password is None:
            # Use Private Key.
            if not private_key:
                # Try to use default key.
                if os.path.exists(os.path.expanduser('~/.ssh/id_rsa')):
                    private_key = '~/.ssh/id_rsa'
                elif os.path.exists(os.path.expanduser('~/.ssh/id_dsa')):
                    private_key = '~/.ssh/id_dsa'
                else:
                    raise CredentialException("No password or key specified.")

            if isinstance(private_key, (AgentKey, RSAKey)):
                # use the paramiko agent or rsa key
                self._tconnect['pkey'] = private_key
            else:
                # isn't a paramiko AgentKey or RSAKey, try to build a
                # key from what we assume is a path to a key
                private_key_file = os.path.expanduser(private_key)
                try:  # try rsa
                    self._tconnect['pkey'] = RSAKey.from_private_key_file(
                        private_key_file, private_key_pass)
                except paramiko.SSHException:   # if it fails, try dss
                    # pylint:disable=r0204
                    self._tconnect['pkey'] = DSSKey.from_private_key_file(
                        private_key_file, private_key_pass)

    def _sftp_connect(self):
        """Establish the SFTP connection."""
        if not self._sftp_live:
            self._sftp = paramiko.SFTPClient.from_transport(self._transport)
            self._sftp_live = True

    @property
    def pwd(self):
        '''return the current working directory

        :returns: (str) current working directory

        '''
        self._sftp_connect()
        return self._sftp.normalize('.')

    @contextmanager
    def cd(self, remotepath=None):  # pylint:disable=c0103
        """context manager that can change to a optionally specified remote
        directory and restores the old pwd on exit.

        :param str|None remotepath: *Default: None* -
            remotepath to temporarily make the current directory
        :returns: None
        :raises: IOError, if remote path doesn't exist
        """
        original_path = self.pwd
        try:
            if remotepath is not None:
                self.chdir(remotepath)
            yield
        finally:
            self.chdir(original_path)

    def chdir(self, remotepath):
        """change the current working directory on the remote

        :param str remotepath: the remote path to change to

        :returns: None

        :raises: IOError, if path does not exist

        """
        self._sftp_connect()
        self._sftp.chdir(remotepath)

    def listdir(self, remotepath='.'):
        """return a list of files/directories for the given remote path.
        Unlike, paramiko, the directory listing is sorted.

        :param str remotepath: path to list on the server

        :returns: (list of str) directory entries, sorted

        """
        self._sftp_connect()
        return sorted(self._sftp.listdir(remotepath))

    def isfile(self, remotepath):
        """return true if remotepath is a file

        :param str remotepath: the path to test

        :returns: (bool)

        """
        self._sftp_connect()
        try:
            result = S_ISREG(self._sftp.stat(remotepath).st_mode)
        except IOError:     # no such file
            result = False
        return result

    def get(self, remotepath, localpath=None, callback=None,
            preserve_mtime=False):
        """Copies a file between the remote host and the local host.

        :param str remotepath: the remote path and filename, source
        :param str localpath:
            the local path and filename to copy, destination. If not specified,
            file is copied to local current working directory
        :param callable callback:
            optional callback function (form: ``func(int, int)``) that accepts
            the bytes transferred so far and the total bytes to be transferred.
        :param bool preserve_mtime:
            *Default: False* - make the modification time(st_mtime) on the
            local file match the time on the remote. (st_atime can differ
            because stat'ing the localfile can/does update it's st_atime)

        :returns: None

        :raises: IOError

        """
        if not localpath:
            localpath = os.path.split(remotepath)[1]

        self._sftp_connect()
        if preserve_mtime:
            sftpattrs = self._sftp.stat(remotepath)

        self._sftp.get(remotepath, localpath, callback=callback)
        if preserve_mtime:
            os.utime(localpath, (sftpattrs.st_atime, sftpattrs.st_mtime))

    def get_d(self, remotedir, localdir, preserve_mtime=False):
        """get the contents of remotedir and write to locadir. (non-recursive)

        :param str remotedir: the remote directory to copy from (source)
        :param str localdir: the local directory to copy to (target)
        :param bool preserve_mtime: *Default: False* -
            preserve modification time on files

        :returns: None

        :raises:
        """
        self._sftp_connect()
        with self.cd(remotedir):
            for sattr in self._sftp.listdir_attr('.'):
                if S_ISREG(sattr.st_mode):
                    rname = sattr.filename
                    self.get(rname, reparent(localdir, rname),
                             preserve_mtime=preserve_mtime)

    def get_r(self, remotedir, localdir, preserve_mtime=False):
        """recursively copy remotedir structure to localdir

        :param str remotedir: the remote directory to copy from
        :param str localdir: the local directory to copy to
        :param bool preserve_mtime: *Default: False* -
            preserve modification time on files

        :returns: None

        :raises:

        """
        self._sftp_connect()
        wtcb = WTCallbacks()
        self.walktree(remotedir, wtcb.file_cb, wtcb.dir_cb, wtcb.unk_cb)
        # handle directories we recursed through
        for dname in wtcb.dlist:
            for subdir in path_advance(dname):
                try:
                    os.mkdir(reparent(localdir, subdir))
                    # force result to a list for setter,
                    wtcb.dlist = wtcb.dlist + [subdir, ]
                except OSError:     # dir exists
                    pass

        for fname in wtcb.flist:
            # they may have told us to start down farther, so we may not have
            # recursed through some, ensure local dir structure matches
            head, _ = os.path.split(fname)
            if head not in wtcb.dlist:
                for subdir in path_advance(head):
                    if subdir not in wtcb.dlist and subdir != '.':
                        os.mkdir(reparent(localdir, subdir))
                        wtcb.dlist = wtcb.dlist + [subdir, ]

            self.get(fname,
                     reparent(localdir, fname),
                     preserve_mtime=preserve_mtime)

    def walktree(self, remotepath, fcallback, dcallback, ucallback,
                 recurse=True):
        '''recursively descend, depth first, the directory tree rooted at
        remotepath, calling discreet callback functions for each regular file,
        directory and unknown file type.

        :param str remotepath:
            root of remote directory to descend, use '.' to start at
            :attr:`.pwd`
        :param callable fcallback:
            callback function to invoke for a regular file.
            (form: ``func(str)``)
        :param callable dcallback:
            callback function to invoke for a directory. (form: ``func(str)``)
        :param callable ucallback:
            callback function to invoke for an unknown file type.
            (form: ``func(str)``)
        :param bool recurse: *Default: True* - should it recurse

        :returns: None

        :raises:

        '''
        self._sftp_connect()
        for entry in self.listdir(remotepath):
            pathname = posixpath.join(remotepath, entry)
            mode = self._sftp.stat(pathname).st_mode
            if S_ISDIR(mode):
                # It's a directory, call the dcallback function
                dcallback(pathname)
                if recurse:
                    # now, recurse into it
                    self.walktree(pathname, fcallback, dcallback, ucallback)
            elif S_ISREG(mode):
                # It's a file, call the fcallback function
                fcallback(pathname)
            else:
                # Unknown file type
                ucallback(pathname)

    def close(self):
        """Closes the connection and cleans up."""
        # Close SFTP Connection.
        if self._sftp_live:
            self._sftp.close()
            self._sftp_live = False
        # Close the SSH Transport.
        if self._transport:
            self._transport.close()
            self._transport = None


def path_advance(thepath, sep=os.sep):
    '''generator to iterate over a file path forwards

    :param str thepath: the path to navigate forwards
    :param str sep: *Default: os.sep* - the path separator to use

    :returns: (iter)able of strings

    '''
    # handle a direct path
    pre = ''
    if thepath[0] == sep:
        pre = sep
    curpath = ''
    parts = thepath.split(sep)
    if pre:
        if parts[0]:
            parts[0] = pre + parts[0]
        else:
            parts[1] = pre + parts[1]
    for part in parts:
        curpath = os.path.join(curpath, part)
        if curpath:
            yield curpath



import logging
import os
from typing import Dict, Optional

import pandas as pd
import pyspark.sql as ps

from radarpipeline.common import constants, utils
from radarpipeline.datatypes.data_types import DataType
from radarpipeline.io.abc import DataWriter

from pathlib import Path

logger = logging.getLogger(__name__)


class SparkDataWriter(DataWriter):
    """
    Writes the data to a local directory using pySpark
    """

    features: Dict[str, ps.DataFrame]
    num_files: Optional[int] = None
    compression: str

    def __init__(
        self,
        features: Dict[str, DataType],
        output_dir: str,
        compress: bool = False,
        data_format: str = "csv",
    ) -> None:
        super().__init__(features, output_dir)
        self.data_format = data_format
        if data_format == "csv" or data_format == "pickle":
            self.compression = "gzip" if compress is True else "uncompressed"
        elif data_format == "parquet":
            self.compression = "gzip" if compress is True else "snappy"
        else:
            logger.warning("Invalid data format specified. Using default compression")

    def write_data(self) -> None:
        for feature_name, feature_df in self.features.items():
            if isinstance(feature_df, dict):
                for key, value in feature_df.items():
                    self.save_dataframe(f"{feature_name}_{key}", value)
            else:
                self.save_dataframe(feature_name, feature_df)

    def save_dataframe(self, feature_name, feature_df):
        folder_name = utils.pascal_to_snake_case(feature_name)
        folder_path = os.path.join(self.output_dir, folder_name)
        try:
            if self.data_format == 'csv':
                feature_df.write.csv(
                    path=folder_path,
                    header=True,
                    sep=constants.CSV_DELIMITER,
                    encoding=constants.ENCODING,
                    compression=self.compression,
                    lineSep=constants.LINESEP,
                )
            elif self.data_format == "parquet":
                feature_df.write.parquet(
                    path=folder_path + ".parquet",
                    compression=self.compression,
                )
            else:
                raise ValueError(
                    f"Invalid data format {self.data_format} specified \
                        for spark writer"
                )
        except Exception as e:
            logger.error(
                f"Error writing data to file {folder_path}: {e}", exc_info=True
            )
        logger.info(f"Feature {feature_name} data exported to {folder_path}")


class PandasDataWriter(DataWriter):
    """
    Writes the data to a local directory using pandas
    """

    compression: str
    features: Dict[str, pd.DataFrame]

    def __init__(
        self,
        features: Dict[str, DataType],
        output_dir: str,
        compress: bool = False,
        data_format: str = "csv",
    ) -> None:
        super().__init__(features, output_dir)
        self.data_format = data_format
        if data_format == "csv" or data_format == "pickle":
            self.compression = "gzip" if compress is True else "infer"
        elif data_format == "parquet":
            self.compression = "gzip" if compress is True else "snappy"
        else:
            logger.warning("Invalid data format specified. Using default compression")

    def write_data(self) -> None:
        for feature_name, feature_df in self.features.items():
            if isinstance(feature_df, dict):
                for key, value in feature_df.items():
                    self.save_dataframe(f"{feature_name}_{key}", value)
            else:
                self.save_dataframe(feature_name, feature_df)

    def save_dataframe(self, feature_name, feature_df):
        file_path = utils.get_write_file_attr(feature_name,
                                              self.output_dir,
                                              self.data_format,
                                              self.compression)
        try:
            Path(file_path).parent.mkdir(parents=True, exist_ok=True)
            if self.data_format == "csv":
                feature_df.to_csv(
                    file_path,
                    index=False,
                    sep=constants.CSV_DELIMITER,
                    encoding=constants.ENCODING,
                    compression=self.compression,
                )
            elif self.data_format == "pickle":
                feature_df.to_pickle(file_path, compression=self.compression)
            elif self.data_format == "parquet":
                if self.compression == "infer":
                    feature_df.to_parquet(file_path, compression=None)
                else:
                    feature_df.to_parquet(file_path, compression=self.compression)
            else:
                raise ValueError(
                    f"Invalid data format {self.data_format} specified"
                )
        except Exception as e:
            logger.error(
                f"Error writing data to file {file_path}: {e}", exc_info=True
            )
        logger.info(f"Feature {feature_name} data exported to {file_path}")



from radarpipeline.io.abc import DataReader, SchemaReader
from radarpipeline.io.reader import AvroSchemaReader, Reader, SparkCSVDataReader
from radarpipeline.io.downloader import SftpDataReader
from radarpipeline.io.writer import *
from radarpipeline.io.ingestion import CustomDataReader
from radarpipeline.io.sampler import UserSampler, DataSampler



import logging
import sys
import traceback
from typing import Dict, Union, List
from radarpipeline.project import Project
from radarpipeline.io import CustomDataReader
from radarpipeline.project.validations import ConfigGenerator
import yaml

from radarpipeline.common.logger import logger_init
import requests

logger_init()

logger = logging.getLogger(__name__)


def run(config_path: Union[str, Dict] = "config.yaml"):
    """
    Pipeline entry point.
    config_path could be a local path to a configuration file
    or a remote url to a configuration file.
    """
    try:
        logger.info("Starting the pipeline run...")
        logger.info("Reading and Validating the configuration file...")
        project = Project(input_data=config_path)
        logger.info("Fetching the data...")
        project.read_data()
        logger.info("Computing the features...")
        project.compute_features()
        logger.info("Exporting the features data...")
        project.export_data()
        logger.info("Data exported successfully. Closing Spark Engine")
        project.close_spark_session()
        logger.info("Pipeline run completed successfully")
    except KeyboardInterrupt:
        logger.info("Pipeline run interrupted by user")
        sys.exit(0)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def validate(config: Union[str, Dict]):
    """
    Validate the configuration file.
    """
    try:
        logger.info("Validating the configuration file...")
        project = Project(input_data=config)
        project.validate()
        logger.info("Configuration file is valid")
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def generate_config(config_path: str = "./config.yaml", config_dict: Dict = None):
    """
    Generate a sample configuration file.
    """
    try:
        logger.info("Generating a sample configuration file...")
        generator = ConfigGenerator(config_dict, config_path)
        config = generator.generate_config()
        yaml.dump(config, open(config_path, "w"))
        logger.info(f"Sample configuration file generated at {config_path}")
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def read(source_path: str, variables: Union[str, List[str]]):
    """
    Read data from a source.
    """
    try:
        logger.info("Reading data...")
        input_config = {"source_path": source_path}
        data_reader = CustomDataReader(input_config, variables,
                                       source_type="local",
                                       data_format="csv",
                                       df_type="pandas")
        data = data_reader.read_data()
        logger.info("Data read successfully")
        return data
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def fetch(config: Union[Dict, str]):
    """
    Fetch data from a source.
    """
    try:
        logger.info("Fetching data...")
        project = Project(input_data=config)
        project.fetch_data()
        logger.info("Data fetched successfully")
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def convert_from_dict(config: Dict):
    """
    Convert data from one format to another.
    """
    """
    Pipeline entry point.
    config_path could be a local path to a configuration file
    or a remote url to a configuration file.
    """
    try:
        logger.info("Starting the pipeline run...")
        logger.info("Reading and Validating the configuration file...")
        project = Project(input_data=config)
        logger.info("Fetching the data...")
        project.read_data()
        logger.info("Computing the features...")
        project.compute_features()
        logger.info("Exporting the features data...")
        project.export_data()
        logger.info("Data exported successfully. Closing Spark Engine")
        project.close_spark_session()
        logger.info("Pipeline run completed successfully")
    except KeyboardInterrupt:
        logger.info("Pipeline run interrupted by user")
        sys.exit(0)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def convert(yaml_path: str = None,
            source_path: str = None,
            destination_path: str = None,
            variables: List = None, data_format: str = "csv",
            needs_download=False, download_config=None):
    """_summary_

    Args:
        source_path (str): Path to the source file.
        destination_path (str): Path to the destination file.
        data_format (str, optional): Format of the data. Defaults to "csv".
        needs_download (bool, optional): Whether the data needs to be downloaded.
        Defaults to False.
        download_config (dict, optional): Configuration for downloading the data.
        Defaults to None.
    Reads the data from the source and converts it into a tabular format and saves it.
    """
    try:
        if yaml_path:
            run(yaml_path)
        else:
            if source_path is None:
                logger.info("Source path is required")
                sys.exit(1)
            elif destination_path is None:
                logger.info("Destination path is required")
                sys.exit(1)
            elif variables is None:
                logger.info("Variables are required")
                sys.exit(1)
            elif not isinstance(variables, list):
                if isinstance(variables, str):
                    variables = [variables]
                else:
                    logger.info("Variables should be a list")
                    sys.exit(1)
                if len(variables) == 0:
                    logger.info("Variables should not be empty")
                    sys.exit(1)
            else:
                logger.info("Reading data...")
                config_dict = _generate_tabular_config(source_path,
                                                       destination_path,
                                                       variables,
                                                       data_format,
                                                       needs_download,
                                                       download_config)
                run(config_dict)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def _modify_config(config):
    """
    Modify the input configuration to include the variables of interest
    """
    if "project" not in config:
        config["project"] = {
            "project_name": "custom",
            "description": "custom",
            "version": "custom"
        }
    if "input" not in config:
        raise ValueError("Input configuration is missing")
    if "source_type" not in config["input"]:
        config["input"]["source_type"] = "local"
    if "data_format" not in config["input"]:
        config["input"]["data_format"] = "csv"
    if "output" not in config:
        config["output"] = {
            "output_location": "dataframe",
            "config": {},
            "data_format": "csv",
            "compress": False
        }
    if "configurations" not in config:
        config['configurations'] = {}
        config['configurations']['df_type'] = "pandas"
    return config


def compute_features(input_config: Dict,
                     feature_config: Union[Dict, List[Dict]]):
    """
    Use input configuration and compute features.
    Returns the features as a dictionary.
    Args:
        input_config (Dict): Input configuration dictionary.
        feature_config (Union[Dict, List[Dict]]):
            Feature configuration dictionary or list of dictionaries.
    Returns:
        Dict: Dictionary containing the computed features.
    Raises:
        Exception: If there is an error in fetching the features.
    """
    try:
        logger.info("Getting features...")
        if isinstance(feature_config, dict):
            feature_config = [feature_config]
        config = {
            "input": input_config,
            "features": feature_config
        }
        # Modify the configuration to include the variables of interest
        # and set default values for missing keys
        config = _modify_config(config)
        project = Project(input_data=config)
        project.read_data()
        project.compute_features()
        data = project.features
        logger.info("Features fetched successfully")
        return data
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def _generate_tabular_config(source_path: str,
                             destination_path: str,
                             variables: List, data_format: str,
                             needs_download, download_config):
    """
    Generate a configuration dictionary for converting data to tabular format.
    """
    config_dict = {
        "project": {
            "project_name": "tabularise",
            "description": "custom function to tabularise the data",
            "version": "0.0.0"},
        "input": {
            "source_type": "local",
            "config": {
                "source_path": source_path
            },
            "data_format": "csv"
        },
        "output": {
            "output_location": "local",
            "config": {
                "target_path": destination_path
            },
            "data_format": data_format,
        },
        "configurations": {
            "df_type": "pandas"
        }
    }
    if needs_download:
        config_dict["input"]["download"] = download_config
    config_dict["features"] = [
        {
            "location": "custom",
            "feature_groups": ["Tabularize"],
            "feature_names": [variables]
        }
    ]
    return config_dict


def _gather_clone_urls(organization, no_forks=True):
    """
    gh = pygithub3.Github()
    all_repos = gh.repos.list(user=organization).all()
    for repo in all_repos:

        # Don't print the urls for repos that are forks.
        if no_forks and repo.fork:
            continue

        yield repo.clone_url
    TODO: implement this function
    """


def show_available_pipelines():
    """
    Uses git to show the available pipelines from
    https://github.com/RADAR-base-Analytics/
    """
    try:
        logger.info(
            "Fetching available pipelines from "
            "RADAR-base-Analytics GitHub organization..."
        )
        api_url = "https://api.github.com/orgs/RADAR-base-Analytics/repos"
        response = requests.get(api_url)
        if response.status_code == 200:
            repos = response.json()
            pipelines = [
                {"url": repo["html_url"],
                 "name": repo["name"],
                 "description": repo.get("description", "No description available")}
                for repo in repos if not repo.get("fork", False)]
            logger.info(f"Found {len(pipelines)} available pipelines.")
            return pipelines
        else:
            logger.error(
                f"Failed to fetch repositories: {response.status_code} - "
                f"{response.text}"
            )
            return []
    except Exception as e:
        logger.error(f"An error occurred while fetching pipelines: {e}")
        return []


if __name__ == "__main__":
    run()



import unittest
from radarpipeline.project import Project
from radarpipeline.radarpipeline import radarpipeline
from strictyaml.exceptions import YAMLValidationError
import pathlib as pl
import os
import pandas as pd
from pandas.testing import assert_frame_equal


class TestSampling(unittest.TestCase):

    def setUp(self):
        self.default_config = {
            'project': {
                'project_name': 'mock_project',
                'description': 'mock_description',
                'version': 'mock_version'},
            'input': {
                'source_type': 'mock',
                'config': {'source_path': 'mockdata/mockdata'},
                'data_format': 'csv'
            },
            'configurations': {'df_type': 'pandas'},
            'features': [{
                'location': 'https://github.com/RADAR-base-Analytics/mockfeatures',
                'branch': 'main',
                'feature_groups': ['MockFeatureGroup'],
                'feature_names': [['all']]}],
            'output': {
                'output_location': 'local',
                'config': {'target_path': 'output/mockdata'},
                'data_format': 'csv',
                'compress': False}}

    def get_config_output(self, config):
        project = Project(input_data=config)
        project.read_data()
        project.compute_features()
        output_data = project.features
        project.close_spark_session()
        return output_data

    def test_user_sampling_userid(self):
        user_sampling_config = self.default_config
        user_sampling_config['configurations']['user_sampling'] = {}
        user_sampling_config['configurations']['user_sampling']['method'] = 'userid'
        user_sampling_config['configurations']['user_sampling']['config'] = {
            'userids': ["2a02e53a-951e-4fd0-b47f-195a87096bd0"]}
        output_data = self.get_config_output(user_sampling_config)
        self.assertEqual(output_data['PhoneBatteryChargingDuration'][
            'key.userId'].unique(), ['2a02e53a-951e-4fd0-b47f-195a87096bd0'])
        self.assertEqual(output_data['StepCountPerDay']['key.userId'].unique(),
                         ['2a02e53a-951e-4fd0-b47f-195a87096bd0'])

    def test_user_sampling_count(self):
        user_sampling_config = self.default_config
        user_sampling_config['configurations']['user_sampling'] = {}
        user_sampling_config['configurations']['user_sampling']['method'] = 'count'
        user_sampling_config['configurations']['user_sampling']['config'] = {'count': 2}
        output_data = self.get_config_output(user_sampling_config)
        self.assertEqual(output_data['PhoneBatteryChargingDuration'][
            'key.userId'].unique().shape[0], 2)
        self.assertEqual(output_data['StepCountPerDay'][
            'key.userId'].unique().shape[0], 2)

    def test_user_sampling_fraction(self):
        user_sampling_config = self.default_config
        user_sampling_config['configurations']['user_sampling'] = {}
        user_sampling_config['configurations']['user_sampling']['method'] = 'fraction'
        user_sampling_config['configurations']['user_sampling'][
            'config'] = {'fraction': 0.75}
        output_data = self.get_config_output(user_sampling_config)
        self.assertEqual(output_data['PhoneBatteryChargingDuration'][
            'key.userId'].unique().shape[0], 3)
        self.assertEqual(output_data['StepCountPerDay'][
            'key.userId'].unique().shape[0], 3)

    def test_data_sampling_time(self):
        starttime = "2018-11-25 00:00:00"
        endtime = "2018-11-29 00:00:00"
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'time'
        data_sampling_config['configurations'][
            'data_sampling']['config'] = {
                'starttime': starttime,
                'endtime': endtime,
                'time_column': 'value.time'}
        output_data = self.get_config_output(data_sampling_config)
        self.assertGreaterEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].min(), pd.Timestamp(starttime).date())
        self.assertLessEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].max(), pd.Timestamp(endtime).date())
        self.assertGreaterEqual(output_data['StepCountPerDay'][
            'date'].min(), pd.Timestamp(starttime).date())
        self.assertLessEqual(output_data['StepCountPerDay'][
            'date'].max(), pd.Timestamp(endtime).date())

    def test_data_sampling_time_list(self):
        starttime = "2018-11-25 00:00:00"
        endtime = "2018-11-29 00:00:00"
        time_list = [{"starttime": starttime, "endtime": endtime,
                      "time_column": 'value.time'}]
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'time'
        data_sampling_config['configurations'][
            'data_sampling']['config'] = time_list
        output_data = self.get_config_output(data_sampling_config)
        self.assertGreaterEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].min(), pd.Timestamp(starttime).date())
        self.assertLessEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].max(), pd.Timestamp(endtime).date())
        self.assertGreaterEqual(output_data['StepCountPerDay'][
            'date'].min(), pd.Timestamp(starttime).date())
        self.assertLessEqual(output_data['StepCountPerDay'][
            'date'].max(), pd.Timestamp(endtime).date())

    def test_data_sampling_multiple_time(self):
        starttime_1 = "2018-11-25 00:00:00"
        endtime_1 = "2018-11-29 00:00:00"
        starttime_2 = "2019-01-01 00:00:00"
        endtime_2 = "2019-04-30 00:00:00"
        time_list = [{"starttime": starttime_1, "endtime": endtime_1,
                      "time_column": 'value.time'},
                     {"starttime": starttime_2, "endtime": endtime_2}]
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'time'
        data_sampling_config['configurations'][
            'data_sampling']['config'] = time_list
        output_data = self.get_config_output(data_sampling_config)
        self.assertGreaterEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].min(), pd.Timestamp(starttime_1).date())
        self.assertLessEqual(output_data['PhoneBatteryChargingDuration'][
            'date'].max(), pd.Timestamp(endtime_2).date())
        self.assertGreaterEqual(output_data['StepCountPerDay'][
            'date'].min(), pd.Timestamp(starttime_1).date())
        self.assertLessEqual(output_data['StepCountPerDay'][
            'date'].max(), pd.Timestamp(endtime_2).date())
        # check output_data['PhoneBatteryChargingDuration']['date'] is
        # between the time range
        # starttime_1 and endtime_1 and starttime_2 and endtime_2
        self.assertTrue(all((output_data['PhoneBatteryChargingDuration']['date']
                             >= pd.Timestamp(starttime_1).date())
                            & (output_data['PhoneBatteryChargingDuration']['date']
                               <= pd.Timestamp(endtime_1).date())
                            | (output_data['PhoneBatteryChargingDuration']['date']
                               >= pd.Timestamp(starttime_2).date())
                            & (output_data['PhoneBatteryChargingDuration']['date']
                               <= pd.Timestamp(endtime_2).date())))

        self.assertTrue(all((output_data['StepCountPerDay']['date']
                             >= pd.Timestamp(starttime_1).date())
                            & (output_data['StepCountPerDay']['date']
                               <= pd.Timestamp(endtime_1).date())
                            | (output_data['StepCountPerDay']['date']
                               >= pd.Timestamp(starttime_2).date())
                            & (output_data['StepCountPerDay']['date']
                               <= pd.Timestamp(endtime_2).date())))

    def test_data_sampling_multiple_time_single_starttime(self):
        starttime_1 = "2018-11-25 00:00:00"
        endtime_1 = "2018-11-29 00:00:00"
        starttime_2 = "2019-01-01 00:00:00"
        time_list = [{"starttime": starttime_1, "endtime": endtime_1,
                      "time_column": 'value.time'},
                     {"starttime": starttime_2}]
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'time'
        data_sampling_config['configurations'][
            'data_sampling']['config'] = time_list
        output_data = self.get_config_output(data_sampling_config)
        # check output_data['PhoneBatteryChargingDuration']['date'] is between
        # the time range
        # starttime_1 and endtime_1 and starttime_2
        self.assertTrue(all((output_data['PhoneBatteryChargingDuration']['date']
                             >= pd.Timestamp(starttime_1).date())
                            & (output_data['PhoneBatteryChargingDuration']['date']
                               <= pd.Timestamp(endtime_1).date())
                            | (output_data['PhoneBatteryChargingDuration']['date']
                               >= pd.Timestamp(starttime_2).date())))

        self.assertTrue(all((output_data['StepCountPerDay']['date']
                             >= pd.Timestamp(starttime_1).date())
                            & (output_data['StepCountPerDay']['date']
                               <= pd.Timestamp(endtime_1).date())
                            | (output_data['StepCountPerDay']['date']
                               >= pd.Timestamp(starttime_2).date())))

    def test_data_sampling_multiple_time_single_endtime(self):
        endtime_1 = "2018-11-29 00:00:00"
        starttime_2 = "2019-01-01 00:00:00"
        endtime_2 = "2019-04-30 00:00:00"
        time_list = [{"endtime": endtime_1,
                      "time_column": 'value.time'},
                     {"starttime": starttime_2, "endtime": endtime_2}]
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'time'
        data_sampling_config['configurations'][
            'data_sampling']['config'] = time_list
        output_data = self.get_config_output(data_sampling_config)
        # check output_data['PhoneBatteryChargingDuration']['date'] is between
        # the time range
        # endtime_1 and starttime_2 and endtime_2
        self.assertTrue(all((output_data['PhoneBatteryChargingDuration']['date']
                             <= pd.Timestamp(endtime_1).date())
                            | (output_data['PhoneBatteryChargingDuration']['date']
                               >= pd.Timestamp(starttime_2).date())
                            & (output_data['PhoneBatteryChargingDuration']['date']
                               <= pd.Timestamp(endtime_2).date())))

        self.assertTrue(all((output_data['StepCountPerDay']['date']
                             <= pd.Timestamp(endtime_1).date())
                            | (output_data['StepCountPerDay']['date']
                               >= pd.Timestamp(starttime_2).date())
                            & (output_data['StepCountPerDay']['date']
                               <= pd.Timestamp(endtime_2).date())))

    def test_data_sampling_count(self):
        data_sampling_config = self.default_config
        data_sampling_config['configurations']['data_sampling'] = {}
        data_sampling_config['configurations']['data_sampling']['method'] = 'count'
        data_sampling_config['configurations']['data_sampling']['config'] = {
            'count': 100
        }
        output_data = self.get_config_output(data_sampling_config)
        # check if count of per key.userId is less than or equal to 100
        self.assertTrue(all(output_data['PhoneBatteryChargingDuration'].groupby(
            'key.userId').size() <= 100))
        self.assertTrue(all(output_data['StepCountPerDay'].groupby(
            'key.userId').size() <= 100))

    def test_user_data_sampling(self):
        user_data_sampling_config = self.default_config
        user_data_sampling_config['configurations']['user_sampling'] = {}
        user_data_sampling_config['configurations']['user_sampling'][
            'method'] = 'userid'
        user_data_sampling_config['configurations']['user_sampling'][
            'config'] = {'userids': ["2a02e53a-951e-4fd0-b47f-195a87096bd0"]}
        user_data_sampling_config['configurations']['data_sampling'] = {}
        user_data_sampling_config['configurations']['data_sampling'][
            'method'] = 'count'
        user_data_sampling_config['configurations']['data_sampling'][
            'config'] = {'count': 100}
        output_data = self.get_config_output(user_data_sampling_config)
        self.assertEqual(output_data['PhoneBatteryChargingDuration'][
            'key.userId'].unique(), ['2a02e53a-951e-4fd0-b47f-195a87096bd0'])
        self.assertEqual(output_data['StepCountPerDay'][
            'key.userId'].unique(), ['2a02e53a-951e-4fd0-b47f-195a87096bd0'])
        self.assertTrue(all(output_data[
            'PhoneBatteryChargingDuration'].groupby('key.userId').size() <= 100))
        self.assertTrue(all(output_data[
            'StepCountPerDay'].groupby('key.userId').size() <= 100))



from abc import ABC, abstractmethod
from typing import Any, List


class Data(ABC):
    """
    Abstract class for data.
    """

    _data: Any
    df_type: str

    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_data(self) -> Any:
        pass

    @abstractmethod
    def set_data(self, data: Any) -> None:
        pass

    @abstractmethod
    def get_data_keys(self) -> List[str]:
        pass

    @abstractmethod
    def get_data_size(self) -> int:
        pass

    def _preprocess_data(self) -> None:
        """
        Preprocess the data
        """
        for key in self.get_data_keys():
            self._data[key]._preprocess_data()



import unittest
from radarpipeline.project import Project
import yaml
from radarpipeline.radarpipeline import radarpipeline
from radarpipeline.project.validations import ConfigValidator
from strictyaml.exceptions import YAMLValidationError
import pathlib as pl
import os
import pandas as pd
from pandas.testing import assert_frame_equal
from time import sleep


class TestIntegration(unittest.TestCase):

    def setUp(self):
        self.output_dir = None

    def assertIsFile(self, path):
        if not pl.Path(path).resolve().is_file():
            raise AssertionError("File does not exist: %s" % str(path))

    def test_run(self):
        # Test that the pipeline runs without error
        # This is a very basic test, but it's a start
        # Assert if pipeline is throwring an error
        try:
            radarpipeline.run()
            raised = False
        except Exception:
            raised = True
        self.assertFalse(raised, 'Exception raised')
        project = Project(input_data="config.yaml")
        # Assert if output files are created
        self.output_dir = project.config['output']["config"]['target_path']
        path = pl.Path(os.path.join(self.output_dir,
                                    "phone_battery_charging_duration.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output"
        expected_df = pd.read_csv(os.path.join(expected_output_path,
                                               "phone_battery_charging_duration.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()
        path = pl.Path(os.path.join(self.output_dir, "step_count_per_day.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(os.path.join(expected_output_path,
                                               "step_count_per_day.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True))
        path.unlink()

    def test_multiple_feature_run(self):
        try:
            radarpipeline.run(
                "tests/resources/test_yamls/config_with_multiple_features.yaml")
            raised = False
        except Exception:
            raised = True
        self.assertFalse(raised, 'Exception raised')
        project = Project(
            input_data="tests/resources/test_yamls/config_with_multiple_features.yaml")
        # Assert if output files are created
        self.output_dir = project.config['output']["config"]['target_path']
        path = pl.Path(os.path.join(self.output_dir,
                                    "phone_battery_charging_duration.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output"
        expected_df = pd.read_csv(os.path.join(expected_output_path,
                                               "phone_battery_charging_duration.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()

        path = pl.Path(os.path.join(self.output_dir, "step_count_per_day.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(os.path.join(expected_output_path,
                                               "step_count_per_day.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'date'])
                           .reset_index(drop=True))
        path.unlink()

        path = pl.Path(os.path.join(
            self.output_dir,
            "tabularize_features_android_phone_battery_level.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output/tabular/"
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_battery_level.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()
        path = pl.Path(
            os.path.join(
                self.output_dir,
                "tabularize_features_android_phone_step_count.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_step_count.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True))
        path.unlink()

    def rmtree(self, f: pl.Path):
        if f.is_file():
            f.unlink()
        else:
            for child in f.iterdir():
                self.rmtree(child)
            f.rmdir()

    def test_read(self):
        # Test the read function
        try:
            data = radarpipeline.read("tests/resources/test_data",
                                      ["android_phone_step_count"])
            MOCK_PATH = ("tests/resources/test_data/test_participant/"
                         "android_phone_step_count/0000_11.csv.gz")
            self.mock_df = pd.read_csv(MOCK_PATH)
            self.mock_df['value.time'] = pd.to_datetime(self.mock_df['value.time'],
                                                        unit='s').astype(
                                                            "datetime64[us]")
            self.mock_df['value.timeReceived'] = pd.to_datetime(
                self.mock_df['value.timeReceived'], unit='s').astype("datetime64[us]")
            self.assertIsInstance(data.get_combined_data_by_variable(
                "android_phone_step_count"), pd.DataFrame)
            assert_frame_equal(data.get_combined_data_by_variable(
                "android_phone_step_count"), self.mock_df, check_dtype=False)
        except Exception as e:
            self.fail(f"radarpipeline.read raised an exception: {e}")

    def test_read_single_variable(self):
        # Test the read function
        try:
            data = radarpipeline.read("tests/resources/test_data",
                                      "android_phone_step_count")
            MOCK_PATH = ("tests/resources/test_data/test_participant/"
                         "android_phone_step_count/0000_11.csv.gz")
            self.mock_df = pd.read_csv(MOCK_PATH)
            self.mock_df['value.time'] = pd.to_datetime(self.mock_df['value.time'],
                                                        unit='s').astype(
                                                            "datetime64[us]")
            self.mock_df['value.timeReceived'] = pd.to_datetime(
                self.mock_df['value.timeReceived'], unit='s').astype("datetime64[us]")
            self.assertIsInstance(data.get_combined_data_by_variable(
                "android_phone_step_count"), pd.DataFrame)
            assert_frame_equal(data.get_combined_data_by_variable(
                "android_phone_step_count"), self.mock_df, check_dtype=False)
        except Exception as e:
            self.fail(f"radarpipeline.read raised an exception: {e}")

#    def test_fetch(self):
#        # Test the fetch function
#        try:
#            data = radarpipeline.fetch("tests/resources/sample_data_source")
#            self.assertIsInstance(data, pd.DataFrame)
#        except Exception as e:
#            self.fail(f"radarpipeline.fetch raised an exception: {e}")
#
    def test_generate_config(self):
        # Test the generate_config function
        try:
            radarpipeline.generate_config(
                "tests/resources/sample_config.yaml")
            config = yaml.load(
                open("tests/resources/sample_config.yaml"),
                Loader=yaml.FullLoader)
            self.assertIsInstance(config, dict)
            valid_input_formats = ["csv", "csv.gz"]
            valid_output_formats = ["csv"]
            validator = ConfigValidator(config, valid_input_formats,
                                        valid_output_formats)
            validator.validate()
            # delete the sample config file
            os.remove("tests/resources/sample_config.yaml")
        except YAMLValidationError as e:
            self.fail(
                f"radarpipeline.generate_config raised a YAMLValidationError: {e}")
        except Exception as e:
            self.fail(f"radarpipeline.generate_config raised an exception: {e}")

    def test_convert_yaml(self):
        # Test the convert function
        try:
            radarpipeline.convert(
                "tests/resources/test_yamls/config_with_custom_feature.yaml")
            raised = False
        except Exception:
            raised = True
        self.assertFalse(raised, 'Exception raised')
        project = Project(
            input_data="tests/resources/test_yamls/config_with_custom_feature.yaml")
        # Assert if output files are created
        self.output_dir = project.config['output']["config"]['target_path']
        path = pl.Path(os.path.join(
            self.output_dir,
            "tabularize_features_android_phone_battery_level.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output/tabular"
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_battery_level.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()
        path = pl.Path(
            os.path.join(
                self.output_dir,
                "tabularize_features_android_phone_step_count.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_step_count.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True))
        path.unlink()

    def test_convert(self):
        # Test the convert function
        try:
            radarpipeline.convert(source_path="mockdata/mockdata",
                                  destination_path="output/tabular/",
                                  variables=["android_phone_step_count",
                                             "android_phone_battery_level"])
            raised = False
        except Exception:
            raised = True
        self.assertFalse(raised, 'Exception raised')
        project = Project(
            input_data="tests/resources/test_yamls/config_with_custom_feature.yaml")
        # Assert if output files are created
        self.output_dir = project.config['output']["config"]['target_path']
        path = pl.Path(os.path.join(
            self.output_dir,
            "tabularize_features_android_phone_battery_level.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_output_path = "tests/resources/expected_output/tabular"
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_battery_level.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True), check_datetimelike_compat=True)
        path.unlink()
        path = pl.Path(
            os.path.join(
                self.output_dir,
                "tabularize_features_android_phone_step_count.csv"))
        self.assertIsFile(path)
        # read the file and verify that the output is the same
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_step_count.csv"))
        actual_df = pd.read_csv(path)
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True))
        path.unlink()

    def test_compute_features(self):
        # Test the compute_features function
        input_config={
            "source_type": "local",
            "config": {
                "source_path": "mockdata/mockdata"
            },
            "data_format": "csv"
        }
        feature_config={
            "location": "custom",
            "feature_groups": ["Tabularize"],
            "feature_names": [["android_phone_battery_level"]]
        }
        output = radarpipeline.compute_features(input_config, feature_config)
        actual_df = output['TabularizeFeatures']['android_phone_battery_level']
        expected_output_path = "tests/resources/expected_output/tabular"
        expected_df = pd.read_csv(
            os.path.join(expected_output_path,
                         "tabularize_features_android_phone_battery_level.csv"))
        expected_df['value.time'] = pd.to_datetime(expected_df['value.time']).astype("datetime64[us]")
        expected_df['value.timeReceived'] = pd.to_datetime(
            expected_df['value.timeReceived']).astype("datetime64[us]")
        assert_frame_equal(expected_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True),
                           actual_df.sort_values(['key.userId', 'value.time'])
                           .reset_index(drop=True), check_datetimelike_compat=True)

    def tearDown(self):
        # Delete the output directory after the test
        # check if self.output_dir
        if self.output_dir:
            self.rmtree(pl.Path(self.output_dir))


import logging
import logging.handlers
import os
import pathlib
from datetime import datetime


def logger_init(level: int = logging.INFO) -> None:
    """
    Initialize the logger

    Parameters
    ----------
    level : int, optional
        The level of the logger.
        The default is logging.INFO.
    """

    # Get logger
    logger = logging.getLogger()  # root logger
    logger.setLevel(level)

    # File handler
    path = "logs"
    year = str(datetime.now().year)
    month = str(datetime.now().month)
    filename = datetime.now().strftime("%Y%m%d") + "_logfile.log"
    log_directory = os.path.join(path, year, month)
    log_filename = os.path.join(log_directory, filename)

    pathlib.Path(log_directory).mkdir(parents=True, exist_ok=True)
    file = logging.handlers.TimedRotatingFileHandler(
        log_filename, when="midnight", interval=1
    )
    file_format = logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s")
    file.setLevel(logging.INFO)
    file.setFormatter(file_format)

    # Stream handler
    stream = logging.StreamHandler()
    stream_format = logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s")
    stream.setLevel(logging.INFO)
    stream.setFormatter(stream_format)

    # Adding all handlers to the logs
    logger.addHandler(file)
    logger.addHandler(stream)



#!/usr/bin/env python3
from setuptools import find_packages, setup
import os


def read_file(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        return file.read()


setup(
    name="radarpipeline",
    version="2.2.0",
    license='Apache',
    description="A python feature generation and visualization package use with RADAR-base project data.",
    url="https://github.com/RADAR-base/radarpipeline",
    download_url='https://github.com/RADAR-base/radarpipeline/archive/refs/tags/v2.1.0.tar.gz',
    readme="README.md",
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author="Heet Sankesara, Amos Folarin",
    author_email="heet.sankesara@kcl.ac.uk, amos.folarin@kcl.ac.uk",
    keywords=['mhealth', 'pipeline', 'big-data'],
    packages=find_packages(),
    install_requires=[
        "twine==6.2.0",
        "pyYaml==6.0.1",
        "pandas==2.2.2",
        "numpy==1.26.4",
        "scipy==1.14.0",
        "pyspark[sql]==3.5.1",
        "GitPython>=3.1.41",
        "strictyaml==1.7.3",
        "paramiko==3.4.0",
        "avro==1.11.3",
        "pyspark[sql]==3.5.1",
        "pyarrow==16.1.0"],
    test_suite="tests",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    entry_points={
        "console_scripts": [
            "radarpipeline = bin.cli:main",
        ]
    }
)



import os
import logging
from git.exc import GitCommandError
from git.repo import Repo

from radarpipeline.common import utils

logger = logging.getLogger(__name__)


class ConfigValidator():
    def __init__(self, config, valid_input_formats, valid_output_formats) -> None:
        self.valid_input_formats = valid_input_formats
        self.valid_output_formats = valid_output_formats
        self.config = config

    def validate(self):
        """
        Validates the config to check if all the keys are present
        """

        required_keys = ["project", "input", "features", "output"]
        for key in required_keys:
            if key not in self.config:
                raise ValueError(f"Key not present in the config: {key}")

        self._validate_input()
        self._validate_configurations()
        self._validate_features()
        self._validate_output()

        logger.info("Config file validated successfully")

    def _update_mock_data(self) -> None:
        """
        Updates the mock data submodule of the åproject
        """

        try:
            MOCK_URL = "https://github.com/RADAR-base-Analytics/mockdata"
            cache_dir = os.path.join(
                os.path.expanduser("~"), ".cache", "radarpipeline", "mockdata")
            if not os.path.exists(cache_dir):
                logger.info("Mock data submodule not found. Cloning it...")
                Repo.clone_from(MOCK_URL, cache_dir)
                logger.info("Mock data input cloned")
            else:
                repo = Repo(
                    cache_dir,
                    search_parent_directories=True,
                )
                repo.remotes.origin.pull()

        except IndexError:
            raise ValueError("mockdata submodule not found")

    def _validate_input(self) -> None:
        """
        Validates the input data config
        """

        if self.config["input"]["source_type"] == "sftp":
            sftp_config_keys = [
                "sftp_host",
                "sftp_username",
                "sftp_source_path",
                "sftp_private_key",
            ]
            for key in sftp_config_keys:
                if key not in self.config["input"]["config"]:
                    raise ValueError(f"Key not present in the config: {key}")

        elif self.config["input"]["source_type"] == "local":
            if "source_path" not in self.config["input"]["config"]:
                raise ValueError("Key not present in the config: source_path")
            else:
                # Check if local_directory is absolute path. If not, then set it.
                local_directory = self.config["input"]["config"]["source_path"]
                if isinstance(local_directory, list):
                    local_directory = [
                        self._get_absolute_path(local_path)
                        for local_path in local_directory]
                    for local_path in local_directory:
                        if not os.path.exists(local_path):
                            raise ValueError(f"Path does not exist: {local_path}")
                else:
                    local_directory = utils.get_absolute_path(local_directory)
                    if not os.path.exists(local_directory):
                        raise ValueError(f"Path does not exist: {local_directory}")
                self.config["input"]["local_directory"] = local_directory

            # Raise error if data_format it not valid input formats
            if self.config["input"]["data_format"] not in self.valid_input_formats:
                raise ValueError("Invalid value for key in input: data_format")

        elif self.config["input"]["source_type"] == "mock":
            self._update_mock_data()

        else:
            raise ValueError("Invalid value for the key: data_location")

    def _validate_configurations(self) -> None:
        """
        Validates the configurations config
        """

        if "configurations" not in self.config:
            self.config["configurations"] = {}

        valid_df_types = ["pandas", "spark"]
        valid_user_sampling_methods = ["fraction", "count", "userid"]

        if "df_type" not in self.config["configurations"]:
            self.config["configurations"]["df_type"] = "pandas"
        elif self.config["configurations"]["df_type"] not in valid_df_types:
            raise ValueError("Invalid value for the key: df_type")
        if "user_sampling" in self.config["configurations"]:
            if "method" not in self.config["configurations"]["user_sampling"]:
                raise ValueError("Key not present in the user_sampling config: method")
            elif (
                self.config["configurations"]["user_sampling"]["method"]
                not in valid_user_sampling_methods
            ):
                raise ValueError("Invalid value for the key: method")
            else:
                self.config["configurations"][
                    "user_sampling"] = self._validate_user_sampling_config(
                        self.config["configurations"]["user_sampling"])
        else:
            self.config["configurations"]["user_sampling"] = None

        # Validating data sampling
        valid_data_sampling_methods = ["fraction", "count", "time"]
        if "data_sampling" in self.config["configurations"]:
            if "method" not in self.config["configurations"]["data_sampling"]:
                raise ValueError("Key not present in the data_sampling config: method")
            elif (
                self.config["configurations"]["data_sampling"]["method"]
                not in valid_data_sampling_methods
            ):
                raise ValueError("Invalid value for the key: method")
            else:
                self.config["configurations"][
                    "data_sampling"] = self._validate_data_sampling_config(
                        self.config["configurations"]["data_sampling"])
        else:
            self.config["configurations"]["data_sampling"] = None

    def _validate_user_sampling_config(self, user_sampling_config):
        """
        Validates the user_sampling config
        """

        if user_sampling_config["method"] == "fraction":
            user_sampling_config = self._validate_sampling_fraction(
                user_sampling_config)
        elif user_sampling_config["method"] == "count":
            user_sampling_config = self._validate_sampling_count(
                user_sampling_config)
        elif user_sampling_config["method"] == "userid":
            user_sampling_config = self._validate_sampling_userids(
                user_sampling_config)
        return user_sampling_config

    def _validate_data_sampling_config(self, data_sampling_config):
        if data_sampling_config["method"] == "fraction":
            data_sampling_config = self._validate_sampling_fraction(
                data_sampling_config)
        elif data_sampling_config["method"] == "count":
            data_sampling_config = self._validate_sampling_count(
                data_sampling_config)
        elif data_sampling_config["method"] == "time":
            data_sampling_config = self._validate_sampling_time(
                data_sampling_config)
        return data_sampling_config

    def _validate_features(self) -> None:
        """
        Validates the features config
        """

        if len(self.config["features"]) == 0:
            raise ValueError("features array cannot be empty")

        for index, feature in enumerate(self.config["features"]):
            if "location" not in feature:
                raise ValueError(
                    f"Key not present in the config: location at index {index}"
                )

            if "feature_groups" not in feature:
                raise ValueError(
                    f"Key not present in the config: feature_groups at index {index}"
                )

            if len(feature["feature_groups"]) == 0:
                raise ValueError(
                    f"feature_groups array cannot be empty at index {index}"
                )
            if "feature_names" not in feature:
                raise ValueError(
                    f"Key not present in the config: feature_groups at index {index}"
                )

            if len(feature["feature_names"]) == 0:
                raise ValueError(
                    f"feature_groups array cannot be empty at index {index}"
                )
            feature_location = feature["location"]
            if feature_location == "custom":
                pass
            elif utils.is_valid_github_path(feature_location):
                repo_name = utils.get_repo_name_from_url(feature_location)
                cache_dir = os.path.join(
                    os.path.expanduser("~"), ".cache", "radarpipeline", repo_name
                )

                if not os.path.exists(cache_dir):
                    Repo.clone_from(feature_location, cache_dir)

                repo = Repo(cache_dir)
                repo.git.reset("--hard")
                repo.git.clean("-xdf")

                active_branch_name = repo.active_branch.name
                feature_branch = feature.get("branch", active_branch_name)

                try:
                    repo.git.checkout(feature_branch)
                except GitCommandError:
                    logger.warning(
                        "Branch %s does not exist. Using the %s branch instead.",
                        feature_branch,
                        repo.active_branch.name,
                    )
                    feature_branch = repo.active_branch.name
                    repo.git.checkout(feature_branch)
                repo.remotes.origin.pull(feature_branch)

                feature_location = cache_dir
                logger.info(f"Using feature from cache location: {feature_location}")
            else:
                feature_location = os.path.expanduser(feature_location)
                if not os.path.isdir(feature_location):
                    raise ValueError(f"Invalid feature location: {feature_location}")
                logger.info(f"Using feature from local path: {feature_location}")

            self.config["features"][index]["location"] = feature_location

    def _validate_output(self) -> None:
        """
        Validates the output data config
        """

        if self.config["output"]["output_location"] == "postgres":
            postgres_config_keys = [
                "postgres_host",
                "postgres_username",
                "postgres_database",
                "postgres_password",
                "postgres_table",
            ]
            for key in postgres_config_keys:
                if key not in self.config["output"]['config']:
                    raise ValueError(f"Key not present in the config: {key}")

        elif self.config["output"]["output_location"] == "local":
            if "target_path" not in self.config["output"]['config']:
                raise ValueError("Key not present in the config: target_path")
            else:
                # Check if local_directory is absolute path. If not, then set it.
                local_directory = self.config["output"]['config']["target_path"]
                if not os.path.exists(local_directory):
                    os.makedirs(local_directory, exist_ok=True)
                self.config["output"]['config']["target_path"] = local_directory

            # Raise error if data_format it not valid output formats
            if (
                self.config["output"]["data_format"]
                not in self.valid_output_formats
            ):
                raise ValueError("Invalid value for key in output: data_format")

            if "compress" not in self.config["output"]:
                self.config["output"]["compress"] = False
            if self.config["output"]["compress"] == "true":
                self.config["output"]["compress"] = True
        elif self.config["output"]["output_location"] == "dataframe":
            pass
        else:
            raise ValueError("Key not present in the config: output_location")

    def _validate_sampling_fraction(self, sampling_config):
        if "config" not in sampling_config:
            raise ValueError("Key not present in the user_sampling config: config")
        if "fraction" not in sampling_config["config"]:
            raise ValueError(
                "Key not present in the user_sampling config: fraction"
            )
        # converting fraction to float
        try:
            sampling_config["config"]["fraction"] = float(
                sampling_config["config"]["fraction"]
            )
        except ValueError:
            raise ValueError(
                "Invalid value for the key: fraction. It should be a number"
            )
        if not 0 < sampling_config["config"]["fraction"] <= 1:
            raise ValueError(
                "Invalid value for the key: fraction. It should be between 0 and 1"
            )
        return sampling_config

    def _validate_sampling_count(self, sampling_config):
        if "config" not in sampling_config:
            raise ValueError("Key not present in the user_sampling config: config")
        if "count" not in sampling_config["config"]:
            raise ValueError("Key not present in the user_sampling config: count")
        try:
            sampling_config["config"]["count"] = int(
                sampling_config["config"]["count"]
            )
        except ValueError:
            raise ValueError(
                "Invalid value for the key: count. It should be a number"
            )
        if sampling_config["config"]["count"] <= 0:
            raise ValueError(
                "Invalid value for the key: count. It should be greater than 0"
            )
        return sampling_config

    def _validate_sampling_userids(self, sampling_config):
        if "config" not in sampling_config:
            raise ValueError("Key not present in the user_sampling config: config")
        if "userids" not in sampling_config["config"]:
            raise ValueError("Key not present in the user_sampling config: userids")
        if len(sampling_config["config"]["userids"]) == 0:
            raise ValueError(
                "user_ids array cannot be empty in the user_sampling config"
            )
        # check if userids are in array. If not convert to array
        if not isinstance(sampling_config["config"]["userids"], list):
            sampling_config["config"]["userids"] = [
                sampling_config["config"]["userids"]
            ]
        return sampling_config

    def _validate_sampling_time_instance(self, time_dict: dict):
        if ("starttime" not in time_dict
            ) and (
                "endtime" not in time_dict):
            raise ValueError("Neither startime nor endtime present in the config")
        # check if starttime and endtime are can be converted into time format
        # if so, convert them
        if "starttime" in time_dict:
            time_dict['starttime'] = utils.convert_str_to_time(
                time_dict['starttime'])
        if "endtime" in time_dict:
            time_dict['endtime'] = utils.convert_str_to_time(
                time_dict['endtime'])
        if "time_column" not in time_dict:
            logger.warning("time_column not present in the config. \
                Using default time column: value.time")
            time_dict["time_column"] = "value.time"
        return time_dict

    def _validate_sampling_time(self, sampling_config):
        if "config" not in sampling_config:
            raise ValueError("Key not present in the user_sampling config: config")
        if type(sampling_config["config"]) is list:
            if len(sampling_config["config"]) == 0:
                raise ValueError(
                    "No starttime and endtime present in the config"
                )
            for i, time_dict in enumerate(sampling_config["config"]):
                sampling_config["config"][i] = self._validate_sampling_time_instance(
                    time_dict)
        else:
            sampling_config["config"] = self._validate_sampling_time_instance(
                sampling_config["config"])
        return sampling_config


class ConfigGenerator():
    def __init__(self, config_dict=None,
                 path="./config.yaml") -> None:
        self.config_dict = config_dict

    def generate_config(self):
        """
        Generates a default configuration dictionary based on the provided format.
        """
        self.config = self._get_default_config()
        # change default config based on the config dict
        if self.config_dict:
            for key, value in self.config_dict.items():
                if key in self.config:
                    if isinstance(value, dict) and isinstance(self.config[key], dict):
                        self.config[key].update(value)
                    else:
                        self.config[key] = value
        return self.config

    def _get_default_config(self):
        default_config = {
            "project": {
                "project_name": "mock_project",
                "description": "mock_description",
                "version": "mock_version",
            },
            "input": {
                "source_type": "mock",  # could be mock, local, sftp, s3
                "config": {
                    "source_path": "mockdata/mockdata",  # Default for mock/local
                },
                "data_format": "csv",
            },
            "configurations": {
                "df_type": "pandas",
                # Uncomment and configure the following as needed
                # "user_sampling": {
                #     "method": "fraction",
                #     "config": {
                #         "fraction": 0.8,
                #     },
                # },
                # "data_sampling": {
                #     "method": "time",
                #     "config": [
                #         {
                #             "starttime": "2018-11-22 00:00:00",
                #             "endtime": "2018-11-26 00:00:00",
                #             "time_column": "value.time",
                #         },
                #     ],
                # },
            },
            "features": [
                {
                    "location": "https://github.com/RADAR-base-Analytics/mockfeatures",
                    "branch": "main",
                    "feature_groups": ["MockFeatureGroup"],
                    "feature_names": ["all"],
                },
                # Uncomment and configure the following for custom features
                # {
                #     "location": "custom",
                #     "feature_groups": ["Tabularize"],
                #     "feature_names": [
                #         "android_phone_battery_level",
                #         "android_phone_step_count",
                #     ],
                # },
            ],
            "output": {
                "output_location": "local",  # can be local, postgres, sftp
                "config": {
                    "target_path": "output/mockdata/",
                },
                "data_format": "csv",
                "compress": False,
            },
        }
        return default_config

    def save_config(self):
        pass



import pandas as pd
from radarpipeline.datalib import RadarData
from radarpipeline.features import Feature, FeatureGroup
from typing import List, Tuple
from functools import partial


class Tabularize(FeatureGroup):
    def __init__(self, variables: str | List[str]):
        name = "Tabularize"
        description = "custom feature to tabularize the data"
        feature_instance = partial(TabularizeFeature, variables)
        features = [feature_instance]
        super().__init__(name, description, features)
        self.is_custom = True
        if isinstance(variables, str):
            self.required_input_data = [variables]
        self.required_input_data = variables

    def preprocess(self, data: RadarData) -> RadarData:
        """
        Preprocess the data for each feature in the group.
        """
        return data


class TabularizeFeature(Feature):
    def __init__(self, required_input_data):
        self.name = "TabularizeFeatures"
        self.description = "Tabularize multiple variables"
        self.required_input_data = required_input_data

    def preprocess(self, data: RadarData) -> RadarData:
        """
        Preprocess the data for each feature in the group.
        """
        return data

    def calculate(self, data) -> dict:
        tabular_dict = {}
        for key in self.required_input_data:
            tabular_dict[key] = data.get_combined_data_by_variable(
                key)
        return tabular_dict
